import inspect
import json
import os

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)

_currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def _load_entities_data():
    with open(_currentdir + "/entities.json", "r", encoding="utf-8") as f:
        return json.load(f)


_entities_data = _load_entities_data()


def load_entities():
    for character_info in _entities_data.get("EveCharacter"):
        if character_info.get("alliance_id"):
            alliance, _ = EveAllianceInfo.objects.update_or_create(
                alliance_id=character_info.get("alliance_id"),
                defaults={
                    "alliance_name": character_info.get("alliance_name"),
                    "alliance_ticker": character_info.get("alliance_ticker"),
                    "executor_corp_id": character_info.get("corporation_id"),
                },
            )
        else:
            alliance = None

        corporation, _ = EveCorporationInfo.objects.update_or_create(
            corporation_id=character_info.get("corporation_id"),
            defaults={
                "corporation_name": character_info.get("corporation_name"),
                "corporation_ticker": character_info.get("corporation_ticker"),
                "member_count": 99,
                "alliance": alliance,
            },
        )

        EveCharacter.objects.update_or_create(
            character_id=character_info.get("character_id"),
            defaults={
                "character_name": character_info.get("character_name"),
                "corporation_id": corporation.corporation_id,
                "corporation_name": corporation.corporation_name,
                "corporation_ticker": corporation.corporation_ticker,
                "alliance_id": alliance.alliance_id if alliance else None,
                "alliance_name": alliance.alliance_name if alliance else "",
                "alliance_ticker": alliance.alliance_ticker if alliance else "",
            },
        )



# Generated by Django 4.2.29 on 2026-03-21 21:54

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    replaces = [
        ("miningtaxes", "0001_initial"),
        ("miningtaxes", "0002_all_models"),
        ("miningtaxes", "0003_alter_general_options"),
        ("miningtaxes", "0004_settings_interest_rate_alter_settings_phrase"),
        ("miningtaxes", "0005_oreprices_raw_price_oreprices_refined_price_and_more"),
        ("miningtaxes", "0006_charactertaxcredits_credit_type"),
        ("miningtaxes", "0007_character_life_credits_character_life_taxes"),
        ("miningtaxes", "0008_character_monthly_credits_json_and_more"),
        ("miningtaxes", "0009_stats"),
        ("miningtaxes", "0010_stats_admin_get_all_activity_json"),
        ("miningtaxes", "0011_stats_curmonth_leadergraph"),
        ("miningtaxes", "0012_stats_user_mining_ledger_90day"),
        ("miningtaxes", "0013_refactor_tax_rates"),
        ("miningtaxes", "0014_alter_adminminingobslog_eve_solar_system_and_more"),
    ]

    initial = True

    dependencies = [
        ("eve_sde", "0018_blueprintactivity_blueprintactivityproduct_and_more"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app"),
                    (
                        "admin_access",
                        "Can set tax rates for groups and add accountant characters",
                    ),
                    ("auditor_access", "Can view all registered characters data"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AdminCharacter",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                (
                    "eve_character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="miningtaxes_admin_character",
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AdminMiningObservers",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("obs_id", models.BigIntegerField()),
                ("obs_type", models.CharField(max_length=32)),
                ("name", models.CharField(max_length=32)),
                ("sys_name", models.CharField(max_length=32)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="mining_obs",
                        to="miningtaxes.admincharacter",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Character",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("life_credits", models.FloatField(default=0.0)),
                ("life_taxes", models.FloatField(default=0.0)),
                ("monthly_mining_json", models.JSONField(default=None, null=True)),
                ("monthly_taxes_json", models.JSONField(default=None, null=True)),
                ("monthly_credits_json", models.JSONField(default=None, null=True)),
                (
                    "eve_character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="miningtaxes_character",
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Settings",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "phrase",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=10,
                        verbose_name="Keyword (case insensitive) that must be present in the donation reason to be counted. Leave blank/empty to count all donations regardless of reason.",
                    ),
                ),
                (
                    "interest_rate",
                    models.FloatField(
                        default=5.0,
                        verbose_name="Monthly interest rate (%) if taxes have not been paid",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Stats",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("admin_char_json", models.JSONField(default=None, null=True)),
                ("admin_main_json", models.JSONField(default=None, null=True)),
                ("ore_prices_json", models.JSONField(default=None, null=True)),
                ("admin_mining_by_sys_json", models.JSONField(default=None, null=True)),
                ("admin_tax_revenue_json", models.JSONField(default=None, null=True)),
                ("admin_month_json", models.JSONField(default=None, null=True)),
                ("admin_corp_ledger", models.JSONField(default=None, null=True)),
                (
                    "admin_corp_mining_history",
                    models.JSONField(default=None, null=True),
                ),
                ("leaderboards", models.JSONField(default=None, null=True)),
                (
                    "admin_get_all_activity_json",
                    models.JSONField(default=None, null=True),
                ),
                ("curmonth_leadergraph", models.JSONField(default=None, null=True)),
                ("user_mining_ledger_90day", models.JSONField(default=None, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="OrePrices",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("buy", models.FloatField()),
                ("sell", models.FloatField()),
                ("raw_price", models.FloatField(default=0.0)),
                ("refined_price", models.FloatField(default=0.0)),
                ("taxed_price", models.FloatField(default=0.0)),
                ("tax_rate", models.FloatField(default=10.0)),
                ("updated", models.DateTimeField()),
                (
                    "eve_type",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eve_sde.itemtype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CharacterUpdateStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_success",
                    models.BooleanField(db_index=True, default=None, null=True),
                ),
                ("content_hash_1", models.CharField(default="", max_length=32)),
                ("content_hash_2", models.CharField(default="", max_length=32)),
                ("content_hash_3", models.CharField(default="", max_length=32)),
                ("last_error_message", models.TextField()),
                (
                    "root_task_id",
                    models.CharField(
                        db_index=True,
                        default="",
                        help_text="ID of update_all_characters task that started this update",
                        max_length=36,
                    ),
                ),
                (
                    "parent_task_id",
                    models.CharField(
                        db_index=True,
                        default="",
                        help_text="ID of character_update task that started this update",
                        max_length=36,
                    ),
                ),
                (
                    "started_at",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                (
                    "finished_at",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="update_status_set",
                        to="miningtaxes.character",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterTaxCredits",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField(db_index=True)),
                ("credit", models.FloatField(default=0.0)),
                ("credit_type", models.CharField(default="", max_length=32)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tax_credits",
                        to="miningtaxes.character",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterMiningLedgerEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateField(db_index=True)),
                ("quantity", models.PositiveIntegerField()),
                ("raw_price", models.FloatField(default=0.0)),
                ("refined_price", models.FloatField(default=0.0)),
                ("taxed_value", models.FloatField(default=0.0)),
                ("taxes_owed", models.FloatField(default=0.0)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="mining_ledger",
                        to="miningtaxes.character",
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eve_sde.solarsystem",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eve_sde.itemtype",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AdminMiningObsLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateField(db_index=True)),
                ("miner_id", models.BigIntegerField()),
                ("quantity", models.PositiveIntegerField()),
                ("observer_type", models.CharField(max_length=32)),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eve_sde.solarsystem",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eve_sde.itemtype",
                    ),
                ),
                (
                    "observer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="mining_log",
                        to="miningtaxes.adminminingobservers",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="AdminMiningCorpLedgerEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField(db_index=True)),
                ("taxed_id", models.BigIntegerField()),
                ("amount", models.FloatField(default=0.0)),
                ("reason", models.CharField(max_length=32)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="corp_ledger",
                        to="miningtaxes.admincharacter",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.AddConstraint(
            model_name="characterupdatestatus",
            constraint=models.UniqueConstraint(
                fields=("character",),
                name="functional_pk_miningtaxes_charactersyncstatus",
            ),
        ),
        migrations.AddConstraint(
            model_name="charactertaxcredits",
            constraint=models.UniqueConstraint(
                fields=("character", "date", "credit"),
                name="functional_pk_miningtaxes_charactertaxes",
            ),
        ),
        migrations.AddConstraint(
            model_name="characterminingledgerentry",
            constraint=models.UniqueConstraint(
                fields=("character", "date", "eve_solar_system", "eve_type"),
                name="functional_pk_mt_characterminingledgerentry",
            ),
        ),
        migrations.AddConstraint(
            model_name="adminminingobslog",
            constraint=models.UniqueConstraint(
                fields=("observer", "date", "miner_id", "eve_type"),
                name="functional_pk_mt_adminMiningLog",
            ),
        ),
        migrations.AddConstraint(
            model_name="adminminingobservers",
            constraint=models.UniqueConstraint(
                fields=("obs_id",), name="functional_pk_mt_adminMiningObs"
            ),
        ),
        migrations.AddConstraint(
            model_name="adminminingcorpledgerentry",
            constraint=models.UniqueConstraint(
                fields=("character", "date", "taxed_id"),
                name="functional_pk_mt_admincorpledger",
            ),
        ),
    ]



import inspect
import json
import os

from app_utils.esi_testing import EsiClientStub, EsiEndpoint

_currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
_FILENAME_ESI_TESTDATA = "esi_testdata.json"


def load_test_data():
    with open(f"{_currentdir}/{_FILENAME_ESI_TESTDATA}", "r", encoding="utf-8") as f:
        return json.load(f)


_endpoints = [
    EsiEndpoint("Industry", "GetCharactersCharacterIdMining", "character_id"),
    EsiEndpoint(
        "Industry",
        "GetCorporationCorporationIdMiningObservers",
        "corporation_id",
    ),
    EsiEndpoint(
        "Industry",
        "GetCorporationCorporationIdMiningObserversObserverId",
        ("corporation_id", "observer_id"),
    ),
    EsiEndpoint("Universe", "GetUniverseStructuresStructureId", "structure_id"),
    EsiEndpoint(
        "Wallet",
        "GetCorporationsCorporationIdWalletsDivisionJournal",
        ("corporation_id", "division"),
    ),
]

esi_client_stub = EsiClientStub(load_test_data(), endpoints=_endpoints)
esi_client_error_stub = EsiClientStub(
    load_test_data(), endpoints=_endpoints, http_error=True
)



from eve_sde.models import (
    Constellation,
    ItemGroup,
    ItemType,
    ItemTypeMaterials,
    Region,
    SolarSystem,
)

_GROUPS = {
    18: "Ore Materials",
    450: "Standard Ore",
    465: "Ice",
    468: "Mercoxit",
    711: "Gas",
    1884: "R4 Moon Ore",
    1920: "R8 Moon Ore",
    1921: "R16 Moon Ore",
    1922: "R32 Moon Ore",
    1923: "R64 Moon Ore",
}

_ITEM_TYPES = {
    1230: {"name": "Veldspar", "group_id": 450, "portion_size": 100, "base_price": 5},
    11396: {"name": "Mercoxit", "group_id": 468, "portion_size": 100, "base_price": 20},
    16267: {
        "name": "Blue Ice",
        "group_id": 465,
        "portion_size": 1000,
        "base_price": 15,
    },
    16635: {"name": "Titanium", "group_id": 18, "portion_size": 1, "base_price": 2},
    28695: {
        "name": "Amber Cytoserocin",
        "group_id": 711,
        "portion_size": 1,
        "base_price": 12,
    },
    45492: {"name": "Bitumens", "group_id": 1884, "portion_size": 100, "base_price": 8},
    45496: {
        "name": "Cobaltite",
        "group_id": 1920,
        "portion_size": 100,
        "base_price": 8,
    },
    45501: {
        "name": "Carnotite",
        "group_id": 1921,
        "portion_size": 100,
        "base_price": 8,
    },
    45503: {"name": "Xenotime", "group_id": 1922, "portion_size": 100, "base_price": 8},
    45511: {
        "name": "Monazite",
        "group_id": 1923,
        "portion_size": 100,
        "base_price": 10,
    },
    62586: {
        "name": "Compressed Kernite",
        "group_id": 450,
        "portion_size": 1,
        "base_price": 10,
    },
}

_SOLAR_SYSTEMS = {
    30000142: {"name": "Jita", "security_status": 0.9, "visual_effect": None},
    30002537: {"name": "Amamake", "security_status": 0.4, "visual_effect": None},
    31000005: {"name": "J123456", "security_status": -0.1, "visual_effect": None},
    30010001: {
        "name": "Pochven Prime",
        "security_status": -0.2,
        "visual_effect": "TRIGLAVIAN_HOME",
    },
}


def load_eve_sde():
    region, _ = Region.objects.update_or_create(
        id=10000001, defaults={"name": "Test Region"}
    )
    constellation, _ = Constellation.objects.update_or_create(
        id=20000001,
        defaults={"name": "Test Constellation", "region": region},
    )

    for group_id, name in _GROUPS.items():
        ItemGroup.objects.update_or_create(
            id=group_id,
            defaults={"name": name, "published": True},
        )

    for type_id, data in _ITEM_TYPES.items():
        ItemType.objects.update_or_create(
            id=type_id,
            defaults={
                "name": data["name"],
                "group_id": data["group_id"],
                "portion_size": data["portion_size"],
                "base_price": data["base_price"],
                "published": True,
            },
        )

    for system_id, data in _SOLAR_SYSTEMS.items():
        SolarSystem.objects.update_or_create(
            id=system_id,
            defaults={
                "name": data["name"],
                "security_status": data["security_status"],
                "visual_effect": data["visual_effect"],
                "constellation": constellation,
            },
        )


def add_material(item_type_id: int, material_type_id: int, quantity: int):
    ItemTypeMaterials.objects.update_or_create(
        item_type_id=item_type_id,
        material_item_type_id=material_type_id,
        defaults={"quantity": quantity},
    )



import requests
from celery import chord, shared_task
from eve_sde.models import ItemType, ItemTypeMaterials

from django.contrib.auth.models import Permission
from django.db import Error
from django.http.response import Http404
from django.shortcuts import get_object_or_404
from django.utils import timezone
from esi.errors import TokenError

from allianceauth.eveonline.models import EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.django import users_with_permission

from .app_settings import (
    MININGTAXES_PING_CURRENT_MSG,
    MININGTAXES_PING_CURRENT_THRESHOLD,
    MININGTAXES_PING_FIRST_MSG,
    MININGTAXES_PING_INTEREST_APPLIED,
    MININGTAXES_PING_SECOND_MSG,
    MININGTAXES_PING_THRESHOLD,
    MININGTAXES_PRICE_JANICE_API_KEY,
    MININGTAXES_PRICE_JANICE_BUY,
    MININGTAXES_PRICE_JANICE_SELL,
    MININGTAXES_PRICE_JANICE_TIMING,
    MININGTAXES_PRICE_METHOD,
    MININGTAXES_PRICE_SOURCE_ID,
    MININGTAXES_PRICE_SOURCE_NAME,
    MININGTAXES_TASKS_TIME_LIMIT,
    MININGTAXES_TAX_ONLY_CORP_MOONS,
)
from .helpers import PriceGroups
from .models import (
    AdminCharacter,
    AdminMiningCorpLedgerEntry,
    AdminMiningObsLog,
    Character,
    CharacterMiningLedgerEntry,
    OrePrices,
    Settings,
    Stats,
)

logger = get_extension_logger(__name__)
TASK_DEFAULT_KWARGS = {"time_limit": MININGTAXES_TASKS_TIME_LIMIT, "max_retries": 3}


def calctaxes():
    s = Stats.load()
    return s.calctaxes()


def get_user(cid):
    try:
        c = EveCharacter.objects.get(character_id=cid)
        p = c.character_ownership.user.profile
    except Exception:
        return None
    if p is None or p.main_character is None:
        return None
    found = None
    for c in p.user.character_ownerships.all():
        try:
            payee = get_object_or_404(Character, eve_character_id=c.character.pk)
        except Exception:
            continue
        found = payee
        break
    return found


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def notify_taxes_due(self):
    user2taxes = calctaxes()

    for u in user2taxes.keys():
        if user2taxes[u][0] > MININGTAXES_PING_THRESHOLD:
            title = "Taxes are due!"
            message = MININGTAXES_PING_FIRST_MSG.format(user2taxes[u][0])
            notify(user=u, title=title, message=message, level="INFO")


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def notify_second_taxes_due(self):
    user2taxes = calctaxes()

    for u in user2taxes.keys():
        if user2taxes[u][0] > MININGTAXES_PING_THRESHOLD:
            title = "Taxes are due!"
            message = MININGTAXES_PING_SECOND_MSG.format(user2taxes[u][0])
            notify(user=u, title=title, message=message, level="INFO")


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def notify_current_taxes_threshold(self):
    from django.contrib.auth.models import User

    from .models import Stats

    s = Stats.load()
    arr = s.get_admin_main_json()

    for row in arr:
        if row["balance"] >= MININGTAXES_PING_CURRENT_THRESHOLD:
            try:
                u = User.objects.get(id=row["user"])
            except Exception:
                print(f"could not find user: {row['user']}")
                continue
            title = "Taxes are due!"
            message = MININGTAXES_PING_CURRENT_MSG.format(row["balance"])
            notify(user=u, title=title, message=message, level="INFO")


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def apply_interest(self):
    settings = Settings.load()
    user2taxes = calctaxes()

    for u in user2taxes.keys():
        if user2taxes[u][0] <= 0.01:
            continue
        interest = round(user2taxes[u][0] * settings.interest_rate / 100.0, 2)
        if interest > MININGTAXES_PING_THRESHOLD:
            user2taxes[u][2].give_credit(-1.0 * interest, "interest")
            title = "Taxes are overdue!"
            message = MININGTAXES_PING_INTEREST_APPLIED.format(interest)
            notify(user=u, title=title, message=message, level="WARN")


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def auto_add_chars(self):
    from django.db import transaction
    from esi.models import Token

    scopes = Character.get_esi_scopes()
    tracked_corps = set()

    for achar in AdminCharacter.objects.all():
        tracked_corps.add(achar.eve_character.corporation_id)

    for char in EveCharacter.objects.all():
        # check if in plugin already
        try:
            inplug = Character.objects.get(eve_character=char)
            inplug = True
        except Character.DoesNotExist:
            inplug = False
            pass
        if inplug:
            continue
        # check if in tracked corp
        if char.corporation_id not in tracked_corps:
            continue
        # check if has token for esi scope
        token = Token.get_token(char.character_id, scopes)
        if not token:
            continue
        # check if orphaned and find user
        try:
            u = char.character_ownership.user.profile
        except EveCharacter.userprofile.RelatedObjectDoesNotExist:
            continue

        # check to see if MemberAudit is installed and if the character has entries in the mining ledger.
        detectedLedger = None
        try:
            from memberaudit.models import Character as maCharacter

            machar = maCharacter.objects.get(eve_character=char)
            num_ledger = len(machar.mining_ledger.all())
            detectedLedger = False
            if num_ledger > 0:
                detectedLedger = True
        except Exception:
            pass

        if detectedLedger is not None and not detectedLedger:
            continue

        logger.info(f"Need to add {char} to miningtaxes, user: {u}")
        with transaction.atomic():
            mcharacter, _ = Character.objects.update_or_create(eve_character=char)
        update_character(character_pk=mcharacter.pk)


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def update_daily(self):
    logger.info("Beginning price update")
    update_all_prices()

    logger.info("Beginning admin character updates")
    characters = AdminCharacter.objects.all()
    for character in characters:
        update_admin_character(character_pk=character.id, celery=True)

    logger.info("Running character update in parallel")
    chartasks = []
    characters = Character.objects.all()
    for character in characters:
        chartasks.append(update_character.s(character_pk=character.id, celery=True))

    chord(chartasks)(precalcs.s())


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def precalcs(self, output):
    perc = round(sum(output) / len(output) * 100, 2)
    logger.info(f"Starting Precalc: {perc}% update success")
    logger.info("Adding Corp Moon Taxes")
    add_corp_moon_taxes()
    logger.info("Linking Tax Payments")
    add_tax_credits()
    logger.info("Running each character's precalc")
    characters = Character.objects.all()
    # precalc all characters
    for character in characters:
        character.precalc_all()

    logger.info("Running Stats Precalc")
    s = Stats.load()
    s.precalc_all()
    logger.info("Finished precalc, updates complete")


def valid_janice_api_key():
    c = requests.get(
        "https://janice.e-351.com/api/rest/v2/markets",
        headers={
            "Content-Type": "text/plain",
            "X-ApiKey": MININGTAXES_PRICE_JANICE_API_KEY,
            "accept": "application/json",
        },
    ).json()

    if "status" in c:
        logger.debug("Janice API status: %s" % c)
        return False
    else:
        return True


def get_bulk_prices(type_ids):
    r = None
    if MININGTAXES_PRICE_METHOD == "Fuzzwork":
        r = requests.get(
            "https://market.fuzzwork.co.uk/aggregates/",
            params={
                "types": ",".join([str(x) for x in type_ids]),
                "station": MININGTAXES_PRICE_SOURCE_ID,
            },
        ).json()
    elif MININGTAXES_PRICE_METHOD == "Janice":
        r = requests.post(
            "https://janice.e-351.com/api/rest/v2/pricer?market=2",
            data="\n".join([str(x) for x in type_ids]),
            headers={
                "Content-Type": "text/plain",
                "X-ApiKey": MININGTAXES_PRICE_JANICE_API_KEY,
                "accept": "application/json",
            },
        ).json()

        # Make Janice data look like Fuzzworks
        output = {}
        for item in r:
            output[str(item["itemType"]["eid"])] = {
                "buy": {
                    "max": str(
                        item[MININGTAXES_PRICE_JANICE_TIMING][
                            MININGTAXES_PRICE_JANICE_BUY
                        ]
                    )
                },
                "sell": {
                    "min": str(
                        item[MININGTAXES_PRICE_JANICE_TIMING][
                            MININGTAXES_PRICE_JANICE_SELL
                        ]
                    )
                },
            }
        r = output
    else:
        raise f"Unknown pricing method: {MININGTAXES_PRICE_METHOD}"
    return r


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def update_all_prices(self, force=[]):
    type_ids = []
    market_data = {}
    api_up = True
    now = timezone.now()

    # Get all type ids by known groups
    prices = []
    added = set()

    for tid in force:
        if tid in (90665,):  # Patch to ignore bugged dev item for new installs
            continue
        e = ItemType.objects.get(id=tid)
        OrePrices.objects.create(eve_type=e, buy=0, sell=0, updated=now)
        added.add(tid)
        prices.append(e)

    for grp in PriceGroups().groups:
        g = ItemType.objects.filter(published=1, group_id=grp)
        for it in g:
            if it.id in added:
                continue
            added.add(it.id)
            prices.append(it)

    for o in OrePrices.objects.all():
        if o.eve_type_id in added:
            continue
        added.add(o.eve_type_id)
        prices.append(o.eve_type)

    # Update EveUniverse objects
    matset = set()
    for item in prices:
        if item.id in (90665,):  # Patch to ignore bugged dev item
            continue
        materials = ItemTypeMaterials.objects.filter(item_type_id=item.id)
        for mat in materials:
            mat = mat.material_item_type
            matset.add(mat.id)

    if MININGTAXES_PRICE_METHOD == "Fuzzwork":
        logger.debug("Using Fuzzwork")
        logger.debug(
            "Price setup starting for %s items from Fuzzworks API from station id %s (%s), this may take up to 30 seconds..."
            % (
                len(prices),
                MININGTAXES_PRICE_SOURCE_ID,
                MININGTAXES_PRICE_SOURCE_NAME,
            )
        )
    elif MININGTAXES_PRICE_METHOD == "Janice":
        logger.debug("Using Janice")
        if valid_janice_api_key():
            logger.debug(
                "Price setup starting for %s items from Janice API for Jita 4-4, this may take up to 30 seconds..."
                % (len(prices),)
            )
        else:
            logger.debug(
                "Price setup failed for Janice, invalid API key! Provide a working key or change price source to Fuzzwork"
            )
            api_up = False
    else:
        logger.error(
            "Unknown pricing method: '%s', skipping" % MININGTAXES_PRICE_METHOD
        )
        return

    if api_up:
        # Build suitable bulks to fetch prices from API
        for item in prices:
            type_ids.append(item.id)

            if len(type_ids) == 1000:
                market_data.update(get_bulk_prices(type_ids))
                type_ids.clear()

        # Get leftover data from the bulk
        if len(type_ids) > 0:
            market_data.update(get_bulk_prices(type_ids))

        logger.debug("Market data fetched, starting database update...")
        existing = OrePrices.objects.all()
        toupdate = []
        tocreate = []
        for price in prices:
            if not str(price.id) in market_data:
                logger.debug(f"Missing data on {price}")
                continue
            if price.id in matset:
                continue
            buy = int(float(market_data[str(price.id)]["buy"]["max"]))
            sell = int(float(market_data[str(price.id)]["sell"]["min"]))

            found = None
            for e in existing:
                if price.id == e.eve_type.id:
                    found = e
                    break
            if found is not None:
                found.buy = buy
                found.sell = sell
                found.updated = now
                toupdate.append(found)
            else:
                tocreate.append(
                    OrePrices(eve_type_id=price.id, buy=buy, sell=sell, updated=now)
                )

        # Handling refined material prices
        type_ids = list(matset)
        logger.debug(f"Materials price updating: {len(type_ids)}")
        market_data = {}
        market_data.update(get_bulk_prices(type_ids))
        for mat in matset:
            if not str(mat) in market_data:
                logger.debug(f"Missing data on {mat}")
                continue
            buy = int(float(market_data[str(mat)]["buy"]["max"]))
            sell = int(float(market_data[str(mat)]["sell"]["min"]))
            now = timezone.now()

            found = None
            for e in existing:
                if mat == e.eve_type.id:
                    found = e
                    break
            if found is not None:
                found.buy = buy
                found.sell = sell
                found.updated = now
                toupdate.append(found)
            else:
                tocreate.append(
                    OrePrices(eve_type_id=mat, buy=buy, sell=sell, updated=now)
                )

        logger.debug("Objects to be created: %d" % len(tocreate))
        logger.debug("Objects to be updated: %d" % len(toupdate))
        try:
            OrePrices.objects.bulk_create(tocreate)
            OrePrices.objects.bulk_update(toupdate, ["buy", "sell", "updated"])
            logger.debug("All prices succesfully updated")
        except Error as e:
            logger.error("Error updating prices: %s" % e)

        existing = OrePrices.objects.all()
        for e in existing:
            e.calc_prices()
    else:
        logger.error("Price source API is not up! Prices not updated.")


@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def update_admin_character(
    self, character_pk: int, force_update: bool = False, celery=False
) -> bool:
    """Start respective update tasks for all stale sections of a character

    Args:
    - character_pk: PL of character to update
    - force_update: When set to True will always update regardless of stale status

    Returns:
    - True when update was conducted
    - False when no updated was needed
    """
    character = AdminCharacter.objects.get(pk=character_pk)
    if character.is_orphan:
        logger.info("%s: Skipping update for orphaned character", character)
        return False
    needs_update = force_update
    needs_update |= character.is_ledger_stale()

    if not needs_update:
        logger.info("%s: No update required", character)
        return False

    logger.info(
        "%s: Starting %s character update", character, "forced" if force_update else ""
    )

    character.update_all()
    if not celery and MININGTAXES_TAX_ONLY_CORP_MOONS:
        add_corp_moon_taxes()


def add_tax_credits():
    settings = Settings.load()
    characters = AdminCharacter.objects.all()
    phrase = settings.phrase.lower().strip()
    for character in characters:
        entries = character.corp_ledger.all()
        for entry in entries:
            if phrase != "" and phrase not in entry.reason.lower():
                continue
            try:
                payee = get_object_or_404(
                    Character,
                    eve_character_id=EveCharacter.objects.get(
                        character_id=entry.taxed_id
                    ).pk,
                )
            except Character.DoesNotExist:
                payee = get_user(entry.taxed_id)
                if payee is None:
                    continue
                pass
            except Http404:
                continue
            except EveCharacter.DoesNotExist:
                continue
            payee.tax_credits.update_or_create(
                date=entry.date, credit=entry.amount, defaults={"credit_type": "paid"}
            )


def add_tax_credits_by_char(character):
    settings = Settings.load()
    entries = AdminMiningCorpLedgerEntry.objects.filter(
        taxed_id=character.eve_character.character_id
    )
    for entry in entries:
        if settings.phrase != "" and settings.phrase not in entry.reason:
            continue
        character.tax_credits.update_or_create(
            date=entry.date, credit=entry.amount, defaults={"credit_type": "paid"}
        )


def add_corp_moon_taxes():
    characters = Character.objects.all()
    for character in characters:
        add_corp_moon_taxes_by_char(character)


def add_corp_moon_taxes_by_char(character):
    entries = AdminMiningObsLog.objects.filter(
        miner_id=character.eve_character.character_id
    )
    consolidate = {}
    for entry in entries:
        k = f"{entry.date}\t{entry.eve_solar_system_id}\t{entry.eve_type_id}"
        if k not in consolidate:
            consolidate[k] = 0
        consolidate[k] += entry.quantity

    for k in consolidate.keys():
        edate, esys, etype = k.split("\t")
        equantity = consolidate[k]
        try:
            row = character.mining_ledger.get(
                date=edate,
                eve_solar_system_id=esys,
                eve_type_id=etype,
            )
            if row.quantity != equantity:
                row.quantity = equantity
                row.save()
                row.calc_prices(corpmoon=True)
        except CharacterMiningLedgerEntry.DoesNotExist:
            row = character.mining_ledger.create(
                date=edate,
                eve_solar_system_id=esys,
                eve_type_id=etype,
                quantity=equantity,
            )
            row.calc_prices(corpmoon=True)


@shared_task(ignore_result=False, **{**TASK_DEFAULT_KWARGS, **{"bind": True}})
def update_character(
    self, character_pk: int, force_update: bool = False, celery=False
) -> bool:
    """Start respective update tasks for all stale sections of a character

    Args:
    - character_pk: PL of character to update
    - force_update: When set to True will always update regardless of stale status

    Returns:
    - True when update was conducted
    - False when no updated was needed
    """
    character = Character.objects.get(pk=character_pk)
    if character.is_orphan:
        logger.info("%s: Skipping update for orphaned character", character)
        return False
    needs_update = force_update
    needs_update |= character.is_ledger_stale()

    if not needs_update:
        logger.info("%s: No update required", character)
        return False

    try:
        logger.info(
            "%s: Starting %s character update",
            character,
            "forced" if force_update else "",
        )
        character.update_mining_ledger()
        add_tax_credits_by_char(character)
        if not celery and MININGTAXES_TAX_ONLY_CORP_MOONS:
            add_corp_moon_taxes_by_char(character)
    except TokenError:
        p = Permission.objects.get(
            content_type__app_label=Character._meta.app_label, codename="admin_access"
        )
        logger.warn("Missing ESI Token for %s", character)
        title = f"Missing ESI Token for {character}"
        message = f"MiningTaxes could not fetch the mining ledger for {character}."
        for u in users_with_permission(p):
            notify(user=u, title=title, message=message, level="CRITICAL")
        pass
    return False



# Shamelessly stolen from Member Audit
import datetime as dt

from dateutil.relativedelta import relativedelta

from django.db import models
from django.urls import reverse
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.views import bootstrap_icon_plus_name_html

from .. import __title__
from ..helpers import PriceGroups
from . import (
    AdminMiningCorpLedgerEntry,
    AdminMiningObsLog,
    Character,
    CharacterMiningLedgerEntry,
    OrePrices,
    Settings,
    ore_calc_prices,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Stats(models.Model):
    admin_char_json = models.JSONField(default=None, null=True)
    admin_main_json = models.JSONField(default=None, null=True)
    ore_prices_json = models.JSONField(default=None, null=True)
    admin_mining_by_sys_json = models.JSONField(default=None, null=True)
    admin_tax_revenue_json = models.JSONField(default=None, null=True)
    admin_month_json = models.JSONField(default=None, null=True)
    admin_corp_ledger = models.JSONField(default=None, null=True)
    admin_corp_mining_history = models.JSONField(default=None, null=True)
    leaderboards = models.JSONField(default=None, null=True)
    admin_get_all_activity_json = models.JSONField(default=None, null=True)
    curmonth_leadergraph = models.JSONField(default=None, null=True)
    user_mining_ledger_90day = models.JSONField(default=None, null=True)

    def precalc_all(self):
        self.calc_admin_char_json()
        self.calc_admin_main_json()
        self.calc_ore_prices_json()
        self.calc_admin_mining_by_sys_json()
        self.calc_admin_tax_revenue_json()
        self.calc_admin_month_json()
        self.calc_admin_corp_ledger()
        self.calc_admin_corp_mining_history()
        self.calc_leaderboards()
        self.calc_admin_get_all_activity_json()
        self.calc_curmonth_leadergraph()
        self.calc_user_mining_ledger_90day()

    def characterize(self, char):
        eve_char = None
        category = None
        name = None
        try:
            eve_char = EveCharacter.objects.get(character_id=char)
        except EveCharacter.DoesNotExist:
            name = f"<a href='https://evewho.com/character/{char}'>{char}</a>"
            category = "unknown"
            pass
        if eve_char is not None:
            try:
                character = eve_char.miningtaxes_character
                name = character.main_character.character_name
                category = "found"
            except AttributeError:
                name = eve_char.character_name
                category = "unregistered"
                pass
        return name, category

    def calctaxes(self):
        n = now().date()
        curmonth = dt.date(year=n.year, month=n.month, day=1)
        user2taxes = {}
        characters = Character.objects.all()
        for character in characters:
            if character.user is None:
                continue
            taxes = character.get_monthly_taxes()
            total = 0.0
            for k in taxes.keys():
                if k != curmonth:
                    total += taxes[k]
            if character.user not in user2taxes:
                user2taxes[character.user] = [0.0, 0.0, character]
            user2taxes[character.user][0] += total
            if total > user2taxes[character.user][1]:
                user2taxes[character.user][1] = total
                user2taxes[character.user][2] = character
            credits = character.get_lifetime_credits()
            user2taxes[character.user][0] -= credits

        for user in user2taxes.keys():
            taxes_due = round(user2taxes[user][0], 2)
            if taxes_due == 0.00:
                taxes_due = abs(taxes_due)
            user2taxes[user][0] = taxes_due
        return user2taxes

    def calc_admin_char_json(self):
        char_level = {}

        for char in Character.objects.all():
            char_level[char] = {
                "life_tax": char.get_lifetime_taxes(),
                "life_credits": char.get_lifetime_credits(),
            }
            char_level[char]["bal"] = (
                char_level[char]["life_tax"] - char_level[char]["life_credits"]
            )

        char_data = []
        for c in char_level.keys():
            if c.eve_character is None or c.main_character is None:
                continue
            char_data.append(
                {
                    "name": bootstrap_icon_plus_name_html(
                        icon_url=c.eve_character.portrait_url(),
                        name=c.eve_character.character_name,
                        size=16,
                    ),
                    "corp": bootstrap_icon_plus_name_html(
                        icon_url=c.eve_character.corporation_logo_url(),
                        name=c.eve_character.corporation_name,
                        size=16,
                    ),
                    "main_name": bootstrap_icon_plus_name_html(
                        icon_url=c.main_character.portrait_url(),
                        name=c.main_character.character_name,
                        size=16,
                    ),
                    "taxes": char_level[c]["life_tax"],
                    "credits": char_level[c]["life_credits"],
                    "balance": char_level[c]["bal"],
                }
            )
        self.admin_char_json = char_data
        self.save()

    def get_admin_char_json(self):
        if self.admin_char_json is None:
            self.calc_admin_char_json()
        return self.admin_char_json

    def main_data_helper(self, chars):
        main_level = {}
        char2user = {}
        user2taxes = self.calctaxes()

        for char in chars:
            if char.main_character is None:
                logger.error(f"Missing main: {char}")
                continue
            m = char.main_character
            char2user[m] = char.user
            if m not in main_level:
                main_level[m] = {
                    "life_tax": 0.0,
                    "life_credits": 0.0,
                    "last_paid": None,
                }
            main_level[m]["life_tax"] += char.get_lifetime_taxes()
            main_level[m]["life_credits"] += char.get_lifetime_credits()
            if char.last_paid() is not None and (
                main_level[m]["last_paid"] is None
                or char.last_paid() > main_level[m]["last_paid"]
            ):
                main_level[m]["last_paid"] = char.last_paid()
        for m in main_level.keys():
            main_level[m]["balance"] = (
                main_level[m]["life_tax"] - main_level[m]["life_credits"]
            )
        return main_level, char2user, user2taxes

    def calc_admin_main_json(self):
        main_level, char2user, user2taxes = self.main_data_helper(
            Character.objects.all()
        )
        main_data = []
        for i, m in enumerate(main_level.keys()):
            summary_url = reverse("miningtaxes:user_summary", args=[char2user[m].pk])
            action_html = (
                '<a class="btn btn-primary btn-sm" '
                f"href='{summary_url}'>"
                '<i class="fas fa-search"></i></a>'
                '<button type="button" class="btn btn-primary btn-sm" '
                'data-bs-toggle="modal" data-bs-target="#modalCredit" '
                f'onClick="populate({i})" >'
                "$$</button>"
            )
            main_data.append(
                {
                    "name": bootstrap_icon_plus_name_html(
                        icon_url=m.portrait_url(), name=m.character_name, size=16
                    ),
                    "corp": bootstrap_icon_plus_name_html(
                        icon_url=m.corporation_logo_url(),
                        name=m.corporation_name,
                        size=16,
                    ),
                    "balance": main_level[m]["balance"],
                    "last_paid": str(main_level[m]["last_paid"]),
                    "action": action_html,
                    "user": char2user[m].pk,
                    "taxes_due": user2taxes[char2user[m]][0],
                }
            )

        self.admin_main_json = main_data
        self.save()

    def get_admin_main_json(self):
        if self.admin_main_json is None:
            self.calc_admin_main_json()
        return self.admin_main_json

    def calc_ore_prices_json(self):
        data = []
        pg = PriceGroups()
        for ore in OrePrices.objects.all():
            if "Compressed" in ore.eve_type.name:
                continue
            if ore.eve_type.group_id not in pg.taxgroups:
                continue
            raw = 1000.0 * ore.raw_price
            refined = 1000.0 * ore.refined_price
            taxed = 1000.0 * ore.taxed_price
            group = "tax_" + pg.taxgroups[ore.eve_type.group_id]
            tax_rate = ore.tax_rate / 100.0

            tax = taxed * tax_rate
            remaining = taxed - tax
            tax_rate = "{0:.0%}".format(tax_rate)
            group = pg.taxgroups[ore.eve_type.group_id]
            data.append(
                {
                    "group": group,
                    "name": ore.eve_type.name,
                    "raw": raw,
                    "refined": refined,
                    "taxed": taxed,
                    "tax_rate": tax_rate,
                    "remaining": remaining,
                    "tax": tax,
                }
            )
        self.ore_prices_json = data
        self.save()

    def get_ore_prices_json(self):
        if self.ore_prices_json is None:
            self.calc_ore_prices_json()
        return self.ore_prices_json

    def calc_admin_get_all_activity_json(self):
        entries = CharacterMiningLedgerEntry.objects.all().prefetch_related(
            "character", "eve_type", "eve_solar_system"
        )
        sys = {}
        pg = PriceGroups()
        csv_data = [
            [
                "Date",
                "Sys",
                "Character",
                "Main",
                "Ore",
                "Group",
                "Amount",
                "ISK Value",
                "Taxed",
            ]
        ]
        for e in entries:
            try:
                s = e.eve_solar_system.name
                if s not in sys:
                    sys[s] = {}

                group = pg.taxgroups[e.eve_type.group_id]

                csv_data.append(
                    [
                        str(e.date),
                        s,
                        e.character.eve_character.character_name,
                        e.character.main_character.character_name,
                        e.eve_type.name,
                        group,
                        e.quantity,
                        e.taxed_value,
                        e.taxes_owed,
                    ]
                )
            except Exception as e:
                logger.error(f"Failed: {e}")
                continue

        self.admin_get_all_activity_json = csv_data
        self.save()

    def get_admin_get_all_activity_json(self):
        if self.admin_get_all_activity_json is None:
            self.calc_admin_get_all_activity_json()
        return self.admin_get_all_activity_json

    def calc_admin_mining_by_sys_json(self):
        days_90 = now() - dt.timedelta(days=90)
        entries = CharacterMiningLedgerEntry.objects.filter(
            date__gte=days_90
        ).prefetch_related("character", "eve_type", "eve_solar_system")
        sys = {}
        pg = PriceGroups()
        allgroups = set()
        tables = {}

        for e in entries:
            try:
                s = e.eve_solar_system.name
                if s not in sys:
                    sys[s] = {}

                group = pg.taxgroups[e.eve_type.group_id]
                allgroups.add(group)
            except Exception as e:
                logger.error(f"Failed: {e}")
                continue

            month = "%d-%02d" % (e.date.year, e.date.month)
            if month not in tables:
                tables[month] = {}
            if s not in tables[month]:
                tables[month][s] = {}
            if group not in tables[month][s]:
                tables[month][s][group] = {"tax": 0.0, "isk": 0.0}

            tables[month][s][group]["tax"] += e.taxes_owed
            tables[month][s][group]["isk"] += e.taxed_value

            if group not in sys[s]:
                sys[s][group] = {
                    "first": e.date,
                    "last": e.date,
                    "q": e.quantity,
                    "isk": e.taxed_value,
                    "tax": e.taxes_owed,
                }
                continue
            if e.date < sys[s][group]["first"]:
                sys[s][group]["first"] = e.date
            if e.date > sys[s][group]["last"]:
                sys[s][group]["last"] = e.date
            sys[s][group]["isk"] += e.taxed_value
            sys[s][group]["tax"] += e.taxes_owed
            sys[s][group]["q"] += e.quantity

        # Reformat for billboard and calc stats
        for s in sys.keys():
            for g in allgroups:
                if g not in sys[s]:
                    sys[s][g] = {"isk": 0, "tax": 0, "q": 0}
                    continue
                # t = (sys[s][g]["last"] - sys[s][g]["first"]).days
                t = (now().date() - sys[s][g]["first"]).days
                t /= 365.25 / 12
                if t < 1:
                    t = 1
                sys[s][g]["isk"] /= t
                sys[s][g]["tax"] /= t
                sys[s][g]["q"] /= t

        anal = {}
        for a in ("isk", "tax", "q"):
            x = ["x"]
            order = sorted(
                sys.keys(), key=lambda x: (sum(map(lambda y: -sys[x][y][a], allgroups)))
            )
            gorder = sorted(
                allgroups, key=lambda g: sum(map(lambda s: -sys[s][g][a], sys.keys()))
            )
            ys = []
            for g in gorder:
                ys.append([g])
            for s in order:
                x.append(s)
                for i, g in enumerate(gorder):
                    ys[i].append(sys[s][g][a])
                if len(x) > 11:
                    break
            ys.append(x)
            anal[a] = ys
        self.admin_mining_by_sys_json = {
            "anal": anal,
            "tables": tables,
        }
        self.save()

    def get_admin_mining_by_sys_json(self):
        if self.admin_mining_by_sys_json is None:
            self.calc_admin_mining_by_sys_json()
        return self.admin_mining_by_sys_json

    def calc_admin_tax_revenue_json(self):
        entries = AdminMiningCorpLedgerEntry.objects.all()
        settings = Settings.load()

        months = {}
        for e in entries:
            if settings.phrase != "" and settings.phrase not in e.reason:
                continue
            d = dt.date(year=e.date.year, month=e.date.month, day=15)
            if d not in months:
                months[d] = 0.0
            months[d] += e.amount

        xs = list(sorted(months.keys()))
        ys = list(map(lambda x: months[x], xs))
        xs = ["x"] + xs
        ys = ["Revenue"] + ys

        csv_data = [["Month", "Amount (ISK)"]]
        for i in range(1, len(xs)):
            csv_data.append([str(xs[i]), ys[i]])
        xs = list(map(str, xs))
        self.admin_tax_revenue_json = {"xdata": xs, "ydata": ys, "csv": csv_data}
        self.save()

    def get_admin_tax_revenue_json(self):
        if self.admin_tax_revenue_json is None:
            self.calc_admin_tax_revenue_json()
        return self.admin_tax_revenue_json

    def calc_admin_month_json(self):
        characters = Character.objects.all()
        newchars = []
        for c in characters:
            if c.main_character is None:
                logger.error(f"Missing main for {c}")
                continue
            newchars.append(c)
        characters = newchars
        monthly = list(map(lambda x: x.get_monthly_taxes(), characters))
        users = list(map(lambda x: x.main_character.character_name, characters))
        firstmonth = None
        for entries in monthly:
            if len(entries.keys()) == 0:
                continue
            if firstmonth is None or firstmonth > sorted(entries.keys())[0]:
                firstmonth = sorted(entries.keys())[0]
        xs = None
        ys = {}
        for i, entries in enumerate(monthly):
            if not users[i] in ys:
                ys[users[i]] = []
            x = ["x"]
            y = []
            curmonth = firstmonth
            lastmonth = dt.date(now().year, now().month, 1)
            while curmonth <= lastmonth:
                if curmonth not in entries:
                    entries[curmonth] = 0.0
                x.append(curmonth)
                curmonth += relativedelta(months=1)

            if xs is None:
                xs = x

            for yi in range(1, len(xs)):
                y.append(entries[xs[yi]])
            ys[users[i]].append(y)
        yout = []
        for user in sorted(ys.keys()):
            yout.append([user] + [sum(x) for x in zip(*ys[user])])

        yall = ["all"]
        for yi in range(1, len(xs)):
            sumy = 0
            for row in yout:
                sumy += row[yi]
            yall.append(sumy)
        yout.insert(0, yall)
        yout = [yall]  # disable per main retrieval

        csvdata = [["Month", "Main", "Taxes Total"]]
        for xi in range(1, len(xs)):
            month = xs[xi]
            for userarr in yout:
                row = [str(month), userarr[0], userarr[xi]]
                csvdata.append(row)

        xs = list(map(str, xs))
        self.admin_month_json = {"xdata": xs, "ydata": yout, "csv": csvdata}
        self.save()

    def get_admin_month_json(self):
        if self.admin_month_json is None:
            self.calc_admin_month_json()
        return self.admin_month_json

    def calc_admin_corp_ledger(self):
        obs = AdminMiningCorpLedgerEntry.objects.all().order_by("-date")
        data = []
        for o in obs:
            char = o.taxed_id
            name = None
            eve_char = None
            try:
                eve_char = EveCharacter.objects.get(character_id=char)
            except EveCharacter.DoesNotExist:
                name = f"<a href='https://evewho.com/character/{char}'>{char}</a>"
                pass
            except Exception as e:
                logger.error(f"Error unknown user: {char}, error: {e}")
                continue
            if eve_char is not None:
                try:
                    character = eve_char.miningtaxes_character
                    name = character.main_character.character_name
                except AttributeError:
                    name = eve_char.character_name
                    pass
                except Exception as e:
                    logger.error(f"Error unknown user: {eve_char}, error: {e}")
                    continue
            data.append(
                {
                    "date": str(o.date),
                    "name": name,
                    "amount": o.amount,
                    "reason": o.reason,
                }
            )
        self.admin_corp_ledger = {"data": data}
        self.save()

    def get_admin_corp_ledger(self):
        if self.admin_corp_ledger is None:
            self.calc_admin_corp_ledger()
        return self.admin_corp_ledger

    def calc_admin_corp_mining_history(self):
        days_3month = now().date() + relativedelta(months=-3)
        obs = AdminMiningObsLog.objects.filter(date__gte=days_3month).order_by("-date")
        cache = {}
        data = []
        unknown_chars = {}
        unregistered_chars = {}
        for o in obs:
            char = o.miner_id
            if char not in cache:
                cache[char] = self.characterize(char)
            name, category = cache[char]

            if category == "unknown":
                if name not in unknown_chars:
                    unknown_chars[name] = {}
                if o.eve_solar_system not in unknown_chars[name]:
                    unknown_chars[name][o.eve_solar_system] = [0, 0.0, None]
                unknown_chars[name][o.eve_solar_system][0] += o.quantity
                _, _, value = ore_calc_prices(o.eve_type, o.quantity)
                unknown_chars[name][o.eve_solar_system][1] += value
                if (
                    unknown_chars[name][o.eve_solar_system][2] is None
                    or unknown_chars[name][o.eve_solar_system][2] < o.date
                ):
                    unknown_chars[name][o.eve_solar_system][2] = o.date
            elif category == "unregistered":
                if name not in unregistered_chars:
                    unregistered_chars[name] = {}
                if o.eve_solar_system not in unregistered_chars[name]:
                    unregistered_chars[name][o.eve_solar_system] = [0, 0.0, None]
                unregistered_chars[name][o.eve_solar_system][0] += o.quantity
                _, _, value = ore_calc_prices(o.eve_type, o.quantity)
                unregistered_chars[name][o.eve_solar_system][1] += value
                if (
                    unregistered_chars[name][o.eve_solar_system][2] is None
                    or unregistered_chars[name][o.eve_solar_system][2] < o.date
                ):
                    unregistered_chars[name][o.eve_solar_system][2] = o.date
            data.append(
                {
                    "date": str(o.date),
                    "ore": o.eve_type.name,
                    "name": name,
                    "quantity": o.quantity,
                    "location": o.eve_solar_system.name,
                }
            )

        unknown_data = []
        for name in unknown_chars.keys():
            for sys in unknown_chars[name].keys():
                unknown_data.append(
                    {
                        "name": name,
                        "sys": sys.name,
                        "quantity": unknown_chars[name][sys][0],
                        "isk": unknown_chars[name][sys][1],
                        "last": str(unknown_chars[name][sys][2]),
                    }
                )

        unregistered_data = []
        for name in unregistered_chars.keys():
            for sys in unregistered_chars[name].keys():
                unregistered_data.append(
                    {
                        "name": name,
                        "sys": sys.name,
                        "quantity": unregistered_chars[name][sys][0],
                        "isk": unregistered_chars[name][sys][1],
                        "last": str(unregistered_chars[name][sys][2]),
                    }
                )

        self.admin_corp_mining_history = {
            "mining_log": data,
            "unknown_data": unknown_data,
            "unregistered_data": unregistered_data,
        }
        self.save()

    def get_admin_corp_mining_history(self):
        if self.admin_corp_mining_history is None:
            self.calc_admin_corp_mining_history()
        return self.admin_corp_mining_history

    def calc_leaderboards(self):
        characters = Character.objects.all()
        allentries = list(map(lambda x: x.get_monthly_mining(), characters))
        combined = {}
        main2chars = {}
        for i, entries in enumerate(allentries):
            c = characters[i].main_character
            if c is None:
                logger.error(f"Missing main for {c}")
                continue
            if c.character_name not in main2chars:
                main2chars[c.character_name] = []
            main2chars[c.character_name].append(characters[i].pk)
            for m in entries.keys():
                if m not in combined:
                    combined[m] = {}
                if c.character_name not in combined[m]:
                    combined[m][c.character_name] = 0.0
                combined[m][c.character_name] += entries[m]
        output = []
        for m in sorted(combined.keys()):
            users = sorted(combined[m], key=lambda x: -combined[m][x])
            table = []
            for i, u in enumerate(users):
                table.append({"rank": i + 1, "character": u, "amount": combined[m][u]})
            output.append({"month": str(m), "table": table})

        self.leaderboards = {"data": output, "mains": main2chars}
        self.save()

    def calc_curmonth_leadergraph(self):
        lb = self.get_leaderboards()
        curmonth = dt.date(now().year, now().month, 1)
        found = None
        linegraph = {}
        first, last = None, None
        for d in lb["data"]:
            if d["month"] == str(curmonth):
                found = d["table"]
        if found is not None:
            for d in found[0:10]:
                linegraph[d["character"]] = {}
                for cpk in lb["mains"][d["character"]]:
                    c = Character.objects.get(pk=cpk)
                    for dm in c.get_curmonth_daily_mining():
                        dmdate = str(dm.date)
                        if dmdate not in linegraph[d["character"]]:
                            linegraph[d["character"]][dmdate] = 0
                        linegraph[d["character"]][dmdate] += dm.taxed_value
                        if first is None or first > dmdate:
                            first = dmdate
                        if last is None or last < dmdate:
                            last = dmdate
        for c in linegraph.keys():
            final = []
            cumisk = 0
            for d, isk in sorted(linegraph[c].items()):
                cumisk += isk
                final.append({"date": d, "value": cumisk})
            linegraph[c] = final

        for k in linegraph.keys():
            if linegraph[k][0]["date"] != first:
                linegraph[k].insert(0, {"date": first, "value": 0})
            if linegraph[k][-1]["date"] != last:
                linegraph[k].append({"date": last, "value": linegraph[k][-1]["value"]})

        self.curmonth_leadergraph = {"data": linegraph}
        self.save()

    def get_curmonth_leadergraph(self):
        if self.curmonth_leadergraph is None:
            self.calc_curmonth_leadergraph()
        return self.curmonth_leadergraph

    def get_leaderboards(self):
        if self.leaderboards is None:
            self.calc_leaderboards()
        return self.leaderboards

    def calc_user_mining_ledger_90day(self):
        user2chars = {}
        allusers = {}
        pg = PriceGroups()

        for c in Character.objects.all():
            if c.user not in user2chars:
                user2chars[c.user] = []
            user2chars[c.user].append(c)

        for user, characters in user2chars.items():
            if user is None:
                continue
            allpgs = {}
            alldays = {}
            polar = {}
            for c in characters:
                ledger = c.get_90d_mining()
                for entry in ledger:
                    try:
                        g = pg.taxgroups[entry.eve_type.group_id]
                        allpgs[g] = [g]
                        v = entry.taxed_value
                    except Exception as e:
                        logger.warn(f"Unknown entry: {e} - {entry}")
                        continue
                    if entry.date not in alldays:
                        alldays[entry.date] = {}
                    if g not in alldays[entry.date]:
                        alldays[entry.date][g] = 0.0
                    if g not in polar:
                        polar[g] = 0.0
                    alldays[entry.date][g] += v
                    polar[g] += v
            xs = ["x"]
            if len(alldays.keys()) > 0:
                curd = sorted(alldays.keys())[0]
                days = [0, 0]
                while curd <= now().date():
                    xs.append(str(curd))
                    days[1] += 1
                    mined = False
                    for g in allpgs.keys():
                        if curd not in alldays or g not in alldays[curd]:
                            allpgs[g].append(0)
                        else:
                            mined = True
                            allpgs[g].append(alldays[curd][g])
                    if mined:
                        days[0] += 1
                    curd += dt.timedelta(days=1)
                finalgraph = [xs]
                for g in allpgs.keys():
                    finalgraph.append(allpgs[g])

                polargraph = []
                for g in polar.keys():
                    polargraph.append([g, polar[g]])

                days = round(100.0 * days[0] / days[1], 2)
                allusers[user.pk] = {
                    "stacked": finalgraph,
                    "polargraph": polargraph,
                    "days": [["days mined", days]],
                }
        self.user_mining_ledger_90day = allusers
        self.save()

    def get_user_mining_ledger_90day(self):
        if self.user_mining_ledger_90day is None:
            self.calc_user_mining_ledger_90day()
        return self.user_mining_ledger_90day

    def save(self, *args, **kwargs):
        self.pk = 1
        super(Stats, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        pass

    @classmethod
    def load(cls):
        obj, _ = cls.objects.get_or_create(pk=1)
        return obj



# Generated by Django 4.0.7 on 2023-10-13 12:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("miningtaxes", "0011_stats_curmonth_leadergraph"),
    ]

    operations = [
        migrations.AddField(
            model_name="stats",
            name="user_mining_ledger_90day",
            field=models.JSONField(default=None, null=True),
        ),
    ]



import datetime
import hashlib
import json
from unittest.mock import patch

from django.utils.timezone import now
from esi.errors import TokenError

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import NoSocketsTestCase, add_new_token

from ...models import Character, OrePrices
from ..testdata.esi_client_stub import esi_client_stub
from ..testdata.load_entities import load_entities
from ..testdata.load_eve_sde import load_eve_sde
from ..utils import (
    create_character,
    create_character_update_status,
    create_miningtaxes_character,
)

MODELS_PATH = "miningtaxes.models"


class TestCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eve_sde()
        load_entities()

    def test_user_should_return_user_when_not_orphan(self):
        character = create_miningtaxes_character(1001)
        self.assertEqual(
            character.user, character.eve_character.character_ownership.user
        )

    def test_user_should_be_none_when_orphan(self):
        character = create_character(EveCharacter.objects.get(character_id=1121))
        self.assertIsNone(character.user)
        self.assertTrue(character.is_orphan)

    def test_should_return_main_when_it_exists(self):
        character = create_miningtaxes_character(1001)
        self.assertEqual(
            character.main_character,
            character.eve_character.character_ownership.user.profile.main_character,
        )
        self.assertTrue(character.is_main)

    def test_mining_ledger_aggregation(self):
        date = datetime.date(year=2022, month=1, day=15)
        month_n = datetime.date(year=2022, month=1, day=1)
        character = create_miningtaxes_character(1001)
        OrePrices(eve_type_id=45511, buy=10, sell=100, updated=date).calc_prices()
        entry, _ = character.mining_ledger.update_or_create(
            date=date,
            quantity=10,
            eve_type_id=45511,
            eve_solar_system_id=30000142,
        )
        entry.calc_prices()
        monthly = character.get_monthly_taxes()

        self.assertEqual(entry.raw_price, 100)
        self.assertEqual(entry.refined_price, 100)
        self.assertEqual(entry.taxed_value, 100)
        self.assertEqual(entry.taxes_owed, 10)
        self.assertEqual(character.get_lifetime_taxes(), 10)
        self.assertEqual(monthly[month_n], 10)

    def test_tax_credits(self):
        character = create_miningtaxes_character(1001)
        current = now()
        character.give_credit(1234, "paid")
        monthly = character.get_monthly_credits()
        month_n = datetime.date(year=current.year, month=current.month, day=1)

        self.assertEqual(character.get_lifetime_credits(), 1234)
        self.assertEqual(monthly[month_n], 1234)
        self.assertEqual(character.last_paid().date(), current.date())

    def test_fetch_token_raises_for_orphan(self):
        character = create_character(EveCharacter.objects.get(character_id=1121))
        with self.assertRaises(TokenError):
            character.fetch_token()

    def test_fetch_token_returns_matching_token(self):
        character = create_miningtaxes_character(1001)
        fetched_token = character.fetch_token()

        self.assertEqual(
            fetched_token.character_id, character.eve_character.character_id
        )
        self.assertEqual(fetched_token.user, character.user)

    @patch(MODELS_PATH + ".character.esi")
    def test_get_ledger(self, mock_esi):
        mock_esi.client = esi_client_stub
        character = create_miningtaxes_character(1001)
        add_new_token(
            character.user, character.eve_character, Character.get_esi_scopes()
        )

        character.update_mining_ledger()
        ledger = character.mining_ledger.all()

        self.assertEqual(len(ledger), 1)
        entry = ledger[0]
        self.assertEqual(entry.quantity, 4333)
        self.assertEqual(entry.eve_type_id, 62586)
        self.assertEqual(entry.eve_solar_system_id, 30002537)


class TestCharacterUpdateStatus(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eve_sde()
        load_entities()
        cls.character_1001 = create_miningtaxes_character(1001)
        cls.content = {"alpha": 1, "bravo": 2}

    def test_str(self):
        status = create_character_update_status(character=self.character_1001)
        self.assertEqual(str(status), f"{self.character_1001}")

    def test_reset(self):
        status = create_character_update_status(
            character=self.character_1001,
            is_success=True,
            last_error_message="abc",
            root_task_id="a",
            parent_task_id="b",
        )

        status.reset(root_task_id="1", parent_task_id="2")
        status.refresh_from_db()

        self.assertIsNone(status.is_success)
        self.assertEqual(status.last_error_message, "")
        self.assertEqual(status.root_task_id, "1")
        self.assertEqual(status.parent_task_id, "2")

    def test_has_changed(self):
        status = create_character_update_status(
            character=self.character_1001,
            content_hash_1=hashlib.md5(
                json.dumps(self.content).encode("utf-8")
            ).hexdigest(),
        )

        self.assertFalse(status.has_changed(self.content))
        self.assertTrue(status.has_changed({"changed": True}))

    def test_update_content_hash(self):
        status = create_character_update_status(character=self.character_1001)
        status.update_content_hash(self.content, hash_num=2)
        status.refresh_from_db()

        self.assertEqual(
            status.content_hash_2,
            hashlib.md5(json.dumps(self.content).encode("utf-8")).hexdigest(),
        )

    def test_is_updating(self):
        status = create_character_update_status(
            character=self.character_1001, started_at=now(), finished_at=None
        )
        self.assertTrue(status.is_updating)

        status.finished_at = now()
        status.save()
        self.assertFalse(status.is_updating)



import datetime as dt
import inspect
import json
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

from django.http import HttpResponse
from django.test import RequestFactory
from django.urls import reverse
from django.utils.timezone import now

from .. import views
from ..models import Character, OrePrices, Settings
from .base import MiningTaxesBaseTestCase


class TestViews(MiningTaxesBaseTestCase):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.user)
        self.factory = RequestFactory()

    @patch("miningtaxes.views.Stats.load")
    def test_admin_json_views(self, mock_load):
        stats = MagicMock()
        stats.get_admin_char_json.return_value = [{"a": 1}]
        stats.get_admin_main_json.return_value = [{"b": 1}]
        stats.get_admin_get_all_activity_json.return_value = [["h"]]
        stats.get_admin_corp_ledger.return_value = {"data": []}
        stats.get_admin_corp_mining_history.return_value = {"mining_log": []}
        stats.get_admin_mining_by_sys_json.return_value = {"anal": {}}
        stats.get_admin_tax_revenue_json.return_value = {"csv": []}
        stats.get_admin_month_json.return_value = {"csv": []}
        mock_load.return_value = stats

        for name in [
            "admin_char_json",
            "admin_main_json",
            "admin_get_all_activity_json",
            "admin_corp_ledger",
            "admin_corp_mining_history",
            "admin_mining_by_sys_json",
            "admin_tax_revenue_json",
            "admin_month_json",
        ]:
            response = self.client.get(reverse(f"miningtaxes:{name}"))
            self.assertEqual(response.status_code, 200)

    @patch("miningtaxes.views.render", return_value=HttpResponse("ok"))
    def test_admin_launcher_and_tax_table(self, mock_render):
        response = self.client.get(reverse("miningtaxes:admin_launcher"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("miningtaxes:admin_launcher"),
            {"phrase": "abc", "interest_rate": 7},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Settings.load().phrase, "abc")

        response = self.client.get(reverse("miningtaxes:admin_launcher_tax_table"))
        self.assertEqual(response.status_code, 200)
        self.assertIn("data", response.json())

        ore = OrePrices.objects.get(eve_type_id=45511)
        response = self.client.post(
            reverse("miningtaxes:admin_launcher_save_rates"),
            {"tax_data": json.dumps([{"tid": ore.id, "tax_rate": 33}])},
            follow=False,
        )
        self.assertEqual(response.status_code, 302)
        ore.refresh_from_db()
        self.assertEqual(ore.tax_rate, 33)

    @patch("miningtaxes.views.render", return_value=HttpResponse("ok"))
    def test_admin_tables(self, mock_render):
        response = self.client.get(reverse("miningtaxes:admin_tables"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("miningtaxes:admin_tables"),
            {"creditbox": "bad", "userid": self.user.pk},
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("miningtaxes:admin_tables"),
            {"creditbox": "15", "userid": self.user.pk},
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(self.main_character.tax_credits.filter(credit=15).exists())

    @patch("miningtaxes.views.render", return_value=HttpResponse("ok"))
    @patch("miningtaxes.views.Stats.load")
    def test_basic_views_and_json(self, mock_load, mock_render):
        stats = MagicMock()
        stats.get_ore_prices_json.return_value = [{"x": 1}]
        stats.get_leaderboards.return_value = {"data": []}
        stats.get_curmonth_leadergraph.return_value = {"data": {}}
        stats.get_user_mining_ledger_90day.return_value = {
            str(self.user.pk): {"days": []}
        }
        stats.main_data_helper.return_value = (
            {self.ownership.character: {"balance": 1, "last_paid": None}},
            {},
            {self.user: [1]},
        )
        mock_load.return_value = stats

        self.assertEqual(
            self.client.get(reverse("miningtaxes:ore_prices")).status_code, 200
        )
        self.assertEqual(
            self.client.get(reverse("miningtaxes:ore_prices_json")).status_code, 200
        )
        self.assertEqual(self.client.get(reverse("miningtaxes:faq")).status_code, 200)
        self.assertEqual(self.client.get(reverse("miningtaxes:index")).status_code, 302)
        self.assertEqual(
            self.client.get(reverse("miningtaxes:leaderboards")).status_code, 200
        )
        self.assertEqual(
            self.client.get(reverse("miningtaxes:curmonthgraph")).status_code, 200
        )

    @patch("miningtaxes.views.render", return_value=HttpResponse("ok"))
    @patch("miningtaxes.views.humanize_number", side_effect=lambda value: str(value))
    def test_user_views(self, mock_humanize, mock_render):
        response = self.client.get(
            reverse("miningtaxes:user_summary", args=[self.user.pk])
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:summary_month_json", args=[self.user.pk])
            ).status_code,
            200,
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:all_tax_credits", args=[self.user.pk])
            ).status_code,
            200,
        )
        self.assertEqual(
            self.client.get(reverse("miningtaxes:launcher")).status_code, 200
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:user_ledger", args=[self.user.pk])
            ).status_code,
            200,
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:user_ledger_data", args=[self.user.pk])
            ).status_code,
            200,
        )
        self.assertEqual(
            self.client.get(
                reverse(
                    "miningtaxes:character_mining_ledger_data",
                    args=[self.main_character.pk],
                )
            ).status_code,
            200,
        )

    def test_forbidden_user_views(self):
        self.client.force_login(self.other_user)
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:user_summary", args=[self.user.pk])
            ).status_code,
            403,
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:summary_month_json", args=[self.user.pk])
            ).status_code,
            403,
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:all_tax_credits", args=[self.user.pk])
            ).status_code,
            403,
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:user_ledger", args=[self.user.pk])
            ).status_code,
            403,
        )
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:user_ledger_data", args=[self.user.pk])
            ).status_code,
            403,
        )
        self.assertEqual(
            self.client.get(
                reverse(
                    "miningtaxes:character_mining_ledger_data",
                    args=[self.main_character.pk],
                )
            ).status_code,
            403,
        )

    def test_mutating_views(self):
        observer = self.admin_character.mining_obs.first()
        old_entry = observer.mining_log.create(
            date=(now() - dt.timedelta(days=100)).date(),
            miner_id=1001,
            eve_type_id=45511,
            quantity=1,
            observer_type="structure",
            eve_solar_system_id=30002537,
        )
        response = self.client.get(
            reverse("miningtaxes:purge_old_corphistory"), follow=False
        )
        self.assertEqual(response.status_code, 302)
        self.assertFalse(type(old_entry).objects.filter(pk=old_entry.pk).exists())

        response = self.client.get(
            reverse(
                "miningtaxes:remove_admin_registered", args=[self.alt_character.pk]
            ),
            follow=False,
        )
        self.assertEqual(response.status_code, 302)
        response = self.client.get(
            reverse(
                "miningtaxes:remove_admin_character", args=[self.admin_character.pk]
            ),
            follow=False,
        )
        self.assertEqual(response.status_code, 302)

        removable = Character.objects.create(
            eve_character=self.unregistered_ownership.character
        )
        response = self.client.get(
            reverse("miningtaxes:remove_character", args=[removable.pk]), follow=False
        )
        self.assertEqual(response.status_code, 302)

    def test_remove_character_forbidden_and_notfound(self):
        self.client.force_login(self.other_user)
        response = self.client.get(
            reverse("miningtaxes:remove_character", args=[self.main_character.pk])
        )
        self.assertEqual(response.status_code, 403)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("miningtaxes:remove_admin_registered", args=[9999])
        )
        self.assertEqual(response.status_code, 404)
        response = self.client.get(
            reverse("miningtaxes:remove_admin_character", args=[9999])
        )
        self.assertEqual(response.status_code, 404)
        response = self.client.get(reverse("miningtaxes:remove_character", args=[9999]))
        self.assertEqual(response.status_code, 404)

    @patch("miningtaxes.views.messages.success")
    @patch("miningtaxes.views.tasks.update_admin_character.delay")
    def test_add_admin_character_direct(self, mock_delay, mock_success):
        request = self.factory.get("/")
        request.user = self.user
        response = inspect.unwrap(views.add_admin_character)(
            request, SimpleNamespace(character_id=1001)
        )
        self.assertEqual(response.status_code, 302)
        mock_delay.assert_called_once()

    @patch("miningtaxes.views.messages.success")
    @patch("miningtaxes.views.Stats.load")
    @patch("miningtaxes.views.tasks.update_character.delay")
    def test_add_character_direct(self, mock_delay, mock_load, mock_success):
        request = self.factory.get("/")
        request.user = self.user
        mock_load.return_value = MagicMock()
        response = inspect.unwrap(views.add_character)(
            request, SimpleNamespace(character_id=1102)
        )
        self.assertEqual(response.status_code, 302)
        mock_delay.assert_called_once()

    @patch("miningtaxes.views.render", return_value=HttpResponse("ok"))
    def test_character_viewer_and_user_mining_ledger_90day(self, mock_render):
        self.assertEqual(
            self.client.get(
                reverse("miningtaxes:character_viewer", args=[self.main_character.pk])
            ).status_code,
            200,
        )
        with patch("miningtaxes.views.Stats.load") as mock_load:
            mock_load.return_value.get_user_mining_ledger_90day.return_value = {
                str(self.user.pk): {"days": []}
            }
            self.assertEqual(
                self.client.get(
                    reverse("miningtaxes:user_mining_ledger_90day", args=[self.user.pk])
                ).status_code,
                200,
            )
            mock_load.return_value.get_user_mining_ledger_90day.return_value = {}
            self.assertEqual(
                self.client.get(
                    reverse("miningtaxes:user_mining_ledger_90day", args=[self.user.pk])
                ).status_code,
                403,
            )



from .admin import (  # noqa: F401
    AdminCharacter,
    AdminMiningCorpLedgerEntry,
    AdminMiningObsLog,
)
from .character import (  # noqa: F401
    Character,
    CharacterMiningLedgerEntry,
    CharacterUpdateStatus,
)
from .general import General  # noqa: F401
from .orePrices import OrePrices, get_price, get_tax, ore_calc_prices  # noqa: F401
from .settings import Settings  # noqa: F401
from .stats import Stats  # noqa: F401



from types import SimpleNamespace

from django.http import HttpResponse
from django.test import RequestFactory

from .. import auth_hooks
from ..decorators import (
    fetch_character_if_allowed,
    fetch_token_for_character,
    fetch_user_if_allowed,
)
from ..forms import SettingsForm
from ..models import Settings
from ..templatetags.settings import analytics
from .base import MiningTaxesBaseTestCase


class TestMisc(MiningTaxesBaseTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()

    def test_fetch_user_if_allowed(self):
        @fetch_user_if_allowed()
        def view(request, user_pk, user):
            return HttpResponse(str(user.pk))

        request = self.factory.get("/")
        request.user = self.user
        self.assertEqual(view(request, self.user.pk).status_code, 200)
        self.assertEqual(view(request, 9999).status_code, 404)

    def test_fetch_character_if_allowed(self):
        @fetch_character_if_allowed()
        def view(request, character_pk, character):
            return HttpResponse(str(character.pk))

        request = self.factory.get("/")
        request.user = self.user
        self.assertEqual(view(request, self.main_character.pk).status_code, 200)
        request.user = self.other_user
        self.assertEqual(view(request, self.main_character.pk).status_code, 403)
        request.user = self.user
        self.assertEqual(view(request, 9999).status_code, 404)

    def test_fetch_token_for_character(self):
        class Dummy:
            def fetch_token(self, scopes=None):
                return SimpleNamespace(character_name="Dummy", pk=1)

        @fetch_token_for_character("scope")
        def method(character, token):
            return token

        self.assertEqual(method(Dummy()).character_name, "Dummy")

    def test_settings_form_and_model(self):
        settings = Settings.load()
        form = SettingsForm(instance=settings)
        self.assertIn("phrase", form.fields)
        settings.phrase = "abc"
        settings.save()
        self.assertEqual(Settings.load().phrase, "abc")
        settings.delete()
        self.assertEqual(Settings.load().phrase, "abc")

    def test_auth_hooks(self):
        menu = auth_hooks.register_menu()
        request = self.factory.get("/")
        request.user = self.user
        self.assertTrue(menu.render(request))
        request.user = self.other_user
        self.assertTrue(menu.render(request))
        self.assertTrue(auth_hooks.register_urls())

    def test_analytics_tag(self):
        self.assertTrue(analytics())



from unittest.mock import patch

from ...models import Settings, Stats
from ..base import MiningTaxesBaseTestCase


class TestStats(MiningTaxesBaseTestCase):
    @patch("miningtaxes.models.stats.bootstrap_icon_plus_name_html")
    def test_stats_full_pipeline(self, mock_icon):
        mock_icon.side_effect = lambda **kwargs: kwargs["name"]
        stats = Stats.load()

        self.assertEqual(stats.characterize(9999)[1], "unknown")
        self.assertEqual(stats.characterize(1102)[1], "unregistered")
        self.assertEqual(stats.characterize(1001)[1], "found")

        taxes = stats.calctaxes()
        self.assertIn(self.user, taxes)
        self.assertIn(self.other_user, taxes)

        stats.precalc_all()

        self.assertTrue(stats.get_admin_char_json())
        self.assertTrue(stats.get_admin_main_json())
        self.assertTrue(stats.get_ore_prices_json())
        self.assertTrue(stats.get_admin_get_all_activity_json())
        self.assertIn("anal", stats.get_admin_mining_by_sys_json())
        self.assertIn("csv", stats.get_admin_tax_revenue_json())
        self.assertIn("xdata", stats.get_admin_month_json())
        self.assertIn("data", stats.get_admin_corp_ledger())
        corp_history = stats.get_admin_corp_mining_history()
        self.assertIn("mining_log", corp_history)
        self.assertIn("unknown_data", corp_history)
        self.assertIn("unregistered_data", corp_history)
        self.assertIn("data", stats.get_leaderboards())
        self.assertIn("data", stats.get_curmonth_leadergraph())
        ledger_90 = stats.get_user_mining_ledger_90day()
        self.assertIn(self.user.pk, ledger_90)

    @patch("miningtaxes.models.stats.bootstrap_icon_plus_name_html")
    def test_stats_singleton_and_phrase_filter(self, mock_icon):
        mock_icon.side_effect = lambda **kwargs: kwargs["name"]
        stats = Stats.load()
        stats.admin_char_json = [{"value": 1}]
        stats.save()
        self.assertEqual(Stats.load().pk, 1)
        self.assertEqual(Stats.load().get_admin_char_json(), [{"value": 1}])

        settings = Settings.load()
        settings.phrase = "moon"
        settings.save()
        stats.calc_admin_tax_revenue_json()
        data = stats.get_admin_tax_revenue_json()
        self.assertEqual(len(data["csv"]), 2)

        stats.delete()
        self.assertTrue(Stats.objects.filter(pk=1).exists())



from unittest.mock import patch

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import NoSocketsTestCase, add_new_token

from ...models import AdminCharacter
from ..testdata.esi_client_stub import esi_client_stub
from ..testdata.load_entities import load_entities
from ..testdata.load_eve_sde import load_eve_sde
from ..utils import create_character, create_miningtaxes_admincharacter

MODELS_PATH = "miningtaxes.models"


class TestAdminCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eve_sde()
        load_entities()

    def test_user_should_return_user_when_not_orphan(self):
        character = create_miningtaxes_admincharacter(1001)
        self.assertEqual(
            character.user, character.eve_character.character_ownership.user
        )

    def test_user_should_be_none_when_orphan(self):
        character = create_character(EveCharacter.objects.get(character_id=1121))
        self.assertIsNone(character.user)

    def test_should_return_main_when_it_exists(self):
        character = create_miningtaxes_admincharacter(1001)
        self.assertEqual(
            character.main_character,
            character.eve_character.character_ownership.user.profile.main_character,
        )

    def test_get_esi_scopes(self):
        self.assertEqual(
            AdminCharacter.get_esi_scopes(),
            [
                "esi-industry.read_corporation_mining.v1",
                "esi-wallet.read_corporation_wallets.v1",
                "esi-universe.read_structures.v1",
            ],
        )

    @patch(MODELS_PATH + ".admin.esi")
    def test_corp_ledger(self, mock_esi):
        mock_esi.client = esi_client_stub
        character = create_miningtaxes_admincharacter(1001)
        add_new_token(
            character.user, character.eve_character, AdminCharacter.get_esi_scopes()
        )

        character.update_all()
        ledger = character.corp_ledger.all()
        self.assertEqual(len(ledger), 1)
        entry = ledger[0]
        self.assertEqual(entry.amount, 987654321)
        self.assertEqual(entry.reason, "")
        self.assertEqual(entry.taxed_id, 1001)

        observers = character.mining_obs.all()
        self.assertEqual(len(observers), 1)
        obs = observers[0]
        self.assertEqual(obs.obs_id, 123456789)
        self.assertEqual(obs.name, "Amamake - Test Structure Alpha")
        self.assertEqual(obs.sys_name, "Amamake")

        mining_ledger = obs.mining_log.all()
        self.assertEqual(len(mining_ledger), 1)
        entry = mining_ledger[0]
        self.assertEqual(entry.miner_id, 1001)
        self.assertEqual(entry.quantity, 100)
        self.assertEqual(entry.eve_type_id, 45511)



import datetime as dt
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

from django.contrib.auth.models import AnonymousUser
from django.db import IntegrityError
from django.http.response import Http404
from django.test import RequestFactory
from django.utils.timezone import now
from esi.errors import TokenError

from allianceauth.eveonline.models import EveCharacter

from .. import tasks, views
from ..models import AdminCharacter, Character, CharacterUpdateStatus, Settings, Stats
from ..models.admin import AdminMiningObservers
from ..models.admin import _esi_value as admin_esi_value
from ..models.character import _esi_value as char_esi_value
from ..models.character import (
    _is_high_sec,
    _is_low_sec,
    _is_null_sec,
    _is_trig_space,
    _is_w_space,
    _rounded_security_status,
)
from .base import MiningTaxesBaseTestCase
from .utils import create_character_update_status


class TestBranchCoverage(MiningTaxesBaseTestCase):
    def test_character_helper_functions(self):
        trig = SimpleNamespace(
            name="Pochven", security_status=-0.2, visual_effect="TRIGLAVIAN_HOME"
        )
        wormhole = SimpleNamespace(
            name="J123456", security_status=-0.9, visual_effect=""
        )
        hisec = SimpleNamespace(name="Amarr", security_status=0.51, visual_effect="")
        lowsec = SimpleNamespace(name="Auner", security_status=0.35, visual_effect="")
        nullsec = SimpleNamespace(name="Null", security_status=-0.1, visual_effect="")
        unknown = SimpleNamespace(
            name="Unknown", security_status=None, visual_effect=""
        )

        self.assertEqual(admin_esi_value({"x": 1}, "x"), 1)
        self.assertEqual(char_esi_value(SimpleNamespace(x=2), "x"), 2)
        self.assertIsNone(_rounded_security_status(unknown))
        self.assertTrue(_is_trig_space(trig))
        self.assertTrue(_is_w_space(wormhole))
        self.assertFalse(_is_w_space(trig))
        self.assertTrue(_is_high_sec(hisec))
        self.assertTrue(_is_low_sec(lowsec))
        self.assertTrue(_is_null_sec(nullsec))
        self.assertFalse(_is_null_sec(wormhole))

    def test_orphan_character_properties_and_fetch_token(self):
        eve = EveCharacter.objects.create(
            character_id=1300,
            character_name="Orphan Annie",
            corporation_id=2999,
            corporation_name="No Corp",
            corporation_ticker="NC",
            alliance_id=None,
            alliance_name="",
            alliance_ticker="",
        )
        orphan = Character.objects.create(eve_character=eve)

        self.assertTrue(orphan.is_orphan)
        self.assertIsNone(orphan.user)
        self.assertIsNone(orphan.main_character)
        self.assertFalse(orphan.is_main)
        self.assertFalse(orphan.user_is_owner(self.user))
        with self.assertRaises(TokenError):
            orphan.fetch_token()

    @patch("miningtaxes.models.character.notify_throttled")
    def test_fetch_token_missing_notifies(self, mock_notify):
        with self.assertRaises(TokenError):
            self.main_character.fetch_token(scopes=["missing.scope"])
        mock_notify.assert_called_once()

    def test_queryset_access_helpers(self):
        self.assertIn(
            self.main_character.eve_character.character_id,
            Character.objects.eve_character_ids(),
        )
        self.assertTrue(
            Character.objects.owned_by_user(self.user)
            .filter(pk=self.main_character.pk)
            .exists()
        )
        self.assertTrue(
            Character.objects.user_has_access(self.user)
            .filter(pk=self.other_character.pk)
            .exists()
        )
        self.assertFalse(
            Character.objects.user_has_access(self.other_user)
            .filter(pk=self.main_character.pk)
            .exists()
        )
        self.assertTrue(self.main_character.user_has_access(self.user))
        self.assertFalse(self.main_character.user_has_access(self.other_user))

    def test_update_status_helpers(self):
        old_update_section = getattr(Character, "UpdateSection", None)
        Character.UpdateSection = SimpleNamespace(choices=[("ledger", "Ledger")])
        self.assertIsNone(self.main_character.is_update_status_ok())
        status = CharacterUpdateStatus.objects.create(
            character=self.main_character,
            is_success=True,
            started_at=now() - dt.timedelta(minutes=10),
            finished_at=now(),
            last_error_message="",
        )
        self.assertTrue(self.main_character.is_update_status_ok())
        self.assertFalse(status.is_updating)
        status.finished_at = None
        status.save()
        self.assertTrue(status.is_updating)
        self.assertTrue(status.has_changed({"a": 1}))
        status.update_content_hash({"a": 1})
        self.assertFalse(status.has_changed({"a": 1}))
        status.update_content_hash({"b": 2}, hash_num=2)
        status.update_content_hash({"c": 3}, hash_num=3)
        status.reset(root_task_id="root", parent_task_id="parent")
        self.assertEqual(status.root_task_id, "root")
        self.assertEqual(status.parent_task_id, "parent")
        if old_update_section is None:
            delattr(Character, "UpdateSection")
        else:
            Character.UpdateSection = old_update_section

    def test_is_update_status_error_and_ledger_stale(self):
        CharacterUpdateStatus.objects.create(
            character=self.main_character,
            is_success=False,
            started_at=now() - dt.timedelta(minutes=10),
            finished_at=now() - dt.timedelta(minutes=5),
            last_error_message="boom",
        )
        self.assertFalse(self.main_character.is_update_status_ok())
        self.assertTrue(self.main_character.is_ledger_stale())

        self.main_character.update_status_set.all().delete()
        create_character_update_status(
            self.main_character,
            started_at=now()
            - Character.update_time_until_stale()
            - dt.timedelta(minutes=1),
            finished_at=now(),
        )
        self.assertTrue(self.main_character.is_ledger_stale())

    def test_character_aggregate_helpers(self):
        self.assertTrue(self.main_character.get_curmonth_daily_mining().exists())
        self.assertGreaterEqual(self.main_character.get_lifetime_taxes(), 0)
        self.assertGreaterEqual(self.main_character.get_lifetime_credits(), 0)
        self.assertTrue(self.main_character.get_monthly_taxes())
        self.assertTrue(self.main_character.get_monthly_credits())
        self.assertTrue(self.main_character.get_monthly_mining())
        self.assertTrue(self.main_character.get_90d_mining().exists())
        self.assertIsNotNone(self.main_character.last_paid())
        with self.assertRaises(Exception):
            self.main_character.give_credit(1, "bad")
        self.main_character.precalc_all()

    def test_character_json_standardize(self):
        months = [
            {"month": dt.datetime(2024, 1, 1), "total": 1},
            {"month": dt.date(2024, 2, 1), "total": 2},
        ]
        standardized = self.main_character.json_standardize(months)
        self.assertEqual(standardized["2024-01-01"], 1)
        self.assertEqual(
            self.main_character.standardize(standardized)[dt.date(2024, 2, 1)], 2
        )

    @patch("miningtaxes.models.character.get_tax", return_value=0.1)
    @patch(
        "miningtaxes.models.character.ore_calc_prices", return_value=(10.0, 8.0, 6.0)
    )
    def test_calc_prices_tax_filters(self, mock_prices, mock_tax):
        ledger = self.main_character.mining_ledger.create(
            date=now().date(),
            quantity=1,
            eve_type_id=45511,
            eve_solar_system_id=30002537,
        )

        with patch(
            "miningtaxes.models.character.MININGTAXES_WHITELIST", ["Somewhere Else"]
        ):
            ledger.calc_prices()
            self.assertEqual(ledger.taxes_owed, 0.0)

        with patch("miningtaxes.models.character.MININGTAXES_WHITELIST", []), patch(
            "miningtaxes.models.character.MININGTAXES_BLACKLIST",
            [ledger.eve_solar_system.name],
        ):
            ledger.calc_prices()
            self.assertEqual(ledger.taxes_owed, 0.0)

        with patch("miningtaxes.models.character.MININGTAXES_BLACKLIST", []):
            ledger.calc_prices(corpmoon=True)
            self.assertEqual(ledger.taxes_owed, 0.6)

    @patch.object(AdminCharacter, "update_corp_ledger")
    @patch.object(AdminCharacter, "update_mining_observers")
    def test_admin_update_all_and_scopes(self, mock_obs, mock_ledger):
        self.assertIn(
            "esi-wallet.read_corporation_wallets.v1", AdminCharacter.get_esi_scopes()
        )
        with patch("miningtaxes.models.admin.MININGTAXES_TAX_ONLY_CORP_MOONS", True):
            self.admin_character.update_all()
            mock_obs.assert_called_once()
        with patch("miningtaxes.models.admin.MININGTAXES_TAX_ONLY_CORP_MOONS", False):
            self.admin_character.update_all()
        self.assertEqual(mock_ledger.call_count, 2)

    @patch.object(
        AdminCharacter,
        "fetch_token",
        return_value=SimpleNamespace(character_name="Admin", pk=1),
    )
    def test_admin_update_mining_observers_and_ledger(self, mock_fetch_token):
        industry = MagicMock()
        universe = MagicMock()
        wallet = MagicMock()
        mock_client = SimpleNamespace(
            Industry=industry, Universe=universe, Wallet=wallet
        )
        with patch(
            "miningtaxes.models.admin.esi", new=SimpleNamespace(client=mock_client)
        ):

            industry.GetCorporationCorporationIdMiningObservers.return_value.results.return_value = [
                {"observer_id": 777, "observer_type": "structure"},
                {"observer_id": 778, "observer_type": "structure"},
            ]
            universe.GetUniverseStructuresStructureId.side_effect = [
                SimpleNamespace(
                    result=lambda use_etag=False: {
                        "name": "A Very Long Structure Name That Gets Trimmed",
                        "solar_system_id": 30002537,
                    }
                ),
                Exception("gone"),
            ]
            industry.GetCorporationCorporationIdMiningObserversObserverId.return_value.results.return_value = [
                {
                    "character_id": 1001,
                    "last_updated": now().date(),
                    "type_id": 45511,
                    "quantity": 3,
                }
            ]
            wallet.GetCorporationsCorporationIdWalletsDivisionJournal.return_value.results.return_value = [
                {
                    "ref_type": "other",
                    "first_party_id": 1,
                    "date": now(),
                    "amount": 1,
                    "reason": "skip",
                },
                {
                    "ref_type": "player_donation",
                    "first_party_id": 1001,
                    "date": now(),
                    "amount": 5,
                    "reason": "abcdefghijklmnopqrstuvwxyz123456789",
                },
            ]

            self.admin_character.update_mining_observers()
            self.admin_character.update_corp_ledger()

        self.assertTrue(AdminMiningObservers.objects.filter(obs_id=777).exists())
        self.assertTrue(self.admin_character.corp_ledger.filter(taxed_id=1001).exists())
        self.assertTrue(
            self.admin_character.corp_ledger.filter(
                taxed_id=1001, amount=5, reason="abcdefghijklmnopqrstuvwxyz123456"
            ).exists()
        )

    @patch.object(
        AdminCharacter,
        "fetch_token",
        return_value=SimpleNamespace(character_name="Admin", pk=1),
    )
    def test_admin_update_mining_observers_integrity_and_bad_system(
        self, mock_fetch_token
    ):
        existing = self.admin_character.mining_obs.create(
            obs_id=900, obs_type="structure", name="Existing", sys_name="A"
        )
        mock_client = SimpleNamespace(Industry=MagicMock(), Universe=MagicMock())
        with patch(
            "miningtaxes.models.admin.esi", new=SimpleNamespace(client=mock_client)
        ), patch(
            "miningtaxes.models.admin.AdminMiningObservers.objects.update_or_create",
            side_effect=[IntegrityError()],
        ):
            mock_client.Industry.GetCorporationCorporationIdMiningObservers.return_value.results.return_value = [
                {"observer_id": 900, "observer_type": "structure"},
                {"observer_id": 901, "observer_type": "structure"},
            ]
            mock_client.Universe.GetUniverseStructuresStructureId.side_effect = [
                SimpleNamespace(
                    result=lambda use_etag=False: {
                        "name": "Existing Structure",
                        "solar_system_id": 30002537,
                    }
                ),
                SimpleNamespace(
                    result=lambda use_etag=False: {
                        "name": "Bad System",
                        "solar_system_id": 99999999,
                    }
                ),
            ]
            mock_client.Industry.GetCorporationCorporationIdMiningObserversObserverId.return_value.results.return_value = (
                []
            )
            self.admin_character.update_mining_observers()
        self.assertTrue(AdminMiningObservers.objects.filter(pk=existing.pk).exists())

    def test_task_helper_branches(self):
        eve = EveCharacter.objects.create(
            character_id=1301,
            character_name="No Profile",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
            alliance_ticker="WYN",
        )
        self.assertIsNone(tasks.get_user(eve.character_id))
        with patch("miningtaxes.tasks.MININGTAXES_PRICE_METHOD", "Bad"):
            with self.assertRaises(TypeError):
                tasks.get_bulk_prices([1])

    def test_add_tax_credit_branches(self):
        settings = Settings.load()
        settings.phrase = "moon"
        settings.save()

        fallback_char = self.main_character
        fallback_char.tax_credits.all().delete()
        self.admin_character.corp_ledger.create(
            date=now(), taxed_id=1400, amount=10, reason="moon taxes"
        )

        with patch(
            "miningtaxes.tasks.EveCharacter.objects.get",
            return_value=SimpleNamespace(pk=999),
        ), patch(
            "miningtaxes.tasks.get_object_or_404", side_effect=Character.DoesNotExist()
        ), patch(
            "miningtaxes.tasks.get_user", return_value=fallback_char
        ):
            tasks.add_tax_credits()

        self.assertTrue(fallback_char.tax_credits.filter(credit=10).exists())

        with patch(
            "miningtaxes.tasks.EveCharacter.objects.get",
            side_effect=EveCharacter.DoesNotExist(),
        ):
            tasks.add_tax_credits()

        with patch(
            "miningtaxes.tasks.EveCharacter.objects.get",
            return_value=SimpleNamespace(pk=999),
        ), patch("miningtaxes.tasks.get_object_or_404", side_effect=Http404()):
            tasks.add_tax_credits()

    @patch(
        "miningtaxes.tasks.OrePrices.objects.bulk_create", side_effect=tasks.Error("db")
    )
    @patch(
        "miningtaxes.tasks.get_bulk_prices",
        return_value={"45511": {"buy": {"max": "1"}, "sell": {"min": "2"}}},
    )
    def test_update_all_prices_error_branch(self, mock_bulk_prices, mock_bulk_create):
        from ..models import OrePrices

        OrePrices.objects.filter(eve_type_id=45511).delete()
        tasks.update_all_prices.run(force=[45511])
        self.assertTrue(mock_bulk_create.called)

    def test_stats_remaining_paths(self):
        stats = Stats.load()
        orphan_eve = EveCharacter.objects.create(
            character_id=1500,
            character_name="No Main",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
            alliance_ticker="WYN",
        )
        orphan = Character.objects.create(eve_character=orphan_eve)
        orphan.mining_ledger.create(
            date=now().date(),
            quantity=1,
            eve_type_id=45511,
            eve_solar_system_id=30002537,
            taxed_value=10,
            taxes_owed=1,
        )

        with patch(
            "miningtaxes.models.stats.bootstrap_icon_plus_name_html",
            side_effect=lambda **kwargs: kwargs["name"],
        ), patch(
            "miningtaxes.models.stats.EveCharacter.objects.get",
            side_effect=Exception("boom"),
        ):
            stats.calc_admin_corp_ledger()
            self.assertIn("data", stats.get_admin_corp_ledger())

        with patch(
            "miningtaxes.models.stats.bootstrap_icon_plus_name_html",
            side_effect=lambda **kwargs: kwargs["name"],
        ):
            stats.calc_admin_month_json()
            stats.calc_leaderboards()
            stats.calc_curmonth_leadergraph()
            stats.calc_user_mining_ledger_90day()
            self.assertIn("data", stats.get_curmonth_leadergraph())
            self.assertIn(self.user.pk, stats.get_user_mining_ledger_90day())

    @patch("miningtaxes.views.render", return_value=SimpleNamespace(status_code=200))
    def test_view_branches(self, mock_render):
        factory = RequestFactory()
        request = factory.get("/")
        request.user = AnonymousUser()
        self.assertEqual(
            views.remove_character(request, 9999).status_code,
            (
                302
                if hasattr(views.remove_character(request, 9999), "status_code")
                else 302
            ),
        )


class TestCoveragePush(MiningTaxesBaseTestCase):
    def test_auth_hook_hidden_and_fetch_user_forbidden(self):

        from allianceauth.tests.auth_utils import AuthUtils

        from .. import auth_hooks
        from ..decorators import fetch_user_if_allowed

        menu = auth_hooks.register_menu()
        factory = RequestFactory()
        plain_user = AuthUtils.create_user("plain-user")

        request = factory.get("/")
        request.user = plain_user
        self.assertEqual(menu.render(request), "")
        self.assertTrue(auth_hooks.register_urls())

        @fetch_user_if_allowed()
        def view(request, user_pk, user):
            return SimpleNamespace(status_code=200)

        request.user = self.user
        self.assertEqual(view(request, plain_user.pk).status_code, 403)

    def test_oreprices_fallback_paths(self):
        from django.core.exceptions import ObjectDoesNotExist

        from ..models import OrePrices
        from ..models.orePrices import get_price, get_tax, ore_calc_prices

        eve_type = SimpleNamespace(id=999, base_price=None, portion_size=100)
        with patch(
            "miningtaxes.models.orePrices.OrePrices.objects.get",
            side_effect=OrePrices.DoesNotExist(),
        ):
            self.assertEqual(get_tax(eve_type), 0.10)

        with patch(
            "miningtaxes.models.orePrices.OrePrices.objects.get",
            side_effect=ObjectDoesNotExist(),
        ):
            self.assertEqual(get_price(eve_type), 0.0)
            eve_type.base_price = 7
            self.assertEqual(get_price(eve_type), 7)

        materials = [
            SimpleNamespace(quantity=None, material_item_type=SimpleNamespace()),
            SimpleNamespace(quantity=10, material_item_type=SimpleNamespace()),
        ]
        with patch(
            "miningtaxes.models.orePrices.OrePrices.objects.get",
            side_effect=OrePrices.DoesNotExist(),
        ), patch(
            "miningtaxes.models.orePrices.ItemTypeMaterials.objects.filter",
            return_value=materials,
        ), patch(
            "miningtaxes.models.orePrices.get_price", side_effect=[10, 1]
        ), patch(
            "miningtaxes.models.orePrices.MININGTAXES_ALWAYS_TAX_REFINED", False
        ):
            raw, refined, taxed = ore_calc_prices(
                SimpleNamespace(id=1, base_price=10, portion_size=100), 2
            )
            self.assertGreater(raw, refined)
            self.assertEqual(taxed, raw)

        ore = OrePrices.objects.get(eve_type_id=45511)
        with patch(
            "miningtaxes.models.orePrices.ItemTypeMaterials.objects.filter",
            return_value=materials,
        ), patch("miningtaxes.models.orePrices.get_price", return_value=1):
            ore.calc_prices()
            self.assertGreaterEqual(ore.taxed_price, ore.refined_price)

    @patch("miningtaxes.tasks.print")
    @patch("miningtaxes.tasks.notify")
    @patch("miningtaxes.tasks.Stats.load")
    def test_task_notification_skip_branches(self, mock_load, mock_notify, mock_print):
        mock_load.return_value.get_admin_main_json.return_value = [
            {"balance": 2000000000, "user": 999999}
        ]
        tasks.notify_current_taxes_threshold.run()
        mock_notify.assert_not_called()
        mock_print.assert_called_once()

        with patch(
            "miningtaxes.tasks.calctaxes",
            return_value={self.user: [0.0, 0.0, self.main_character]},
        ):
            tasks.notify_taxes_due.run()
            tasks.notify_second_taxes_due.run()
            tasks.apply_interest.run()
        mock_notify.assert_not_called()

        settings = Settings.load()
        settings.interest_rate = 10
        settings.save()
        with patch(
            "miningtaxes.tasks.calctaxes",
            return_value={self.user: [0.1, 0.0, self.main_character]},
        ):
            tasks.apply_interest.run()
        mock_notify.assert_not_called()

    def test_add_corp_moon_taxes_updates_existing_row(self):
        row = self.main_character.mining_ledger.create(
            date=self.current_month,
            quantity=1,
            eve_type_id=45511,
            eve_solar_system_id=30002537,
        )
        from ..models import CharacterMiningLedgerEntry

        with patch.object(CharacterMiningLedgerEntry, "calc_prices") as mock_calc:
            tasks.add_corp_moon_taxes_by_char(self.main_character)
        row.refresh_from_db()
        self.assertEqual(row.quantity, 100)
        mock_calc.assert_called_once()

    @patch("miningtaxes.tasks.add_corp_moon_taxes")
    def test_update_admin_character_skip_and_corpmoon(self, mock_add_corp):
        eve = EveCharacter.objects.create(
            character_id=1600,
            character_name="Orphan Admin",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
            alliance_ticker="WYN",
        )
        orphan_admin = AdminCharacter.objects.create(eve_character=eve)
        self.assertFalse(
            tasks.update_admin_character.run(
                character_pk=orphan_admin.pk, force_update=True
            )
        )

        with patch.object(AdminCharacter, "is_ledger_stale", return_value=False):
            self.assertFalse(
                tasks.update_admin_character.run(
                    character_pk=self.admin_character.pk, force_update=False
                )
            )

        with patch.object(AdminCharacter, "update_all") as mock_update, patch(
            "miningtaxes.tasks.MININGTAXES_TAX_ONLY_CORP_MOONS", True
        ):
            tasks.update_admin_character.run(
                character_pk=self.admin_character.pk, force_update=True, celery=False
            )
        mock_update.assert_called_once()
        mock_add_corp.assert_called_once()

    @patch("miningtaxes.tasks.add_corp_moon_taxes_by_char")
    def test_update_character_skip_and_corpmoon(self, mock_add_corp):
        eve = EveCharacter.objects.create(
            character_id=1601,
            character_name="Orphan Char",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
            alliance_ticker="WYN",
        )
        orphan_char = Character.objects.create(eve_character=eve)
        self.assertFalse(
            tasks.update_character.run(character_pk=orphan_char.pk, force_update=True)
        )

        with patch.object(Character, "is_ledger_stale", return_value=False):
            self.assertFalse(
                tasks.update_character.run(
                    character_pk=self.main_character.pk, force_update=False
                )
            )

        with patch.object(Character, "update_mining_ledger") as mock_update, patch(
            "miningtaxes.tasks.add_tax_credits_by_char"
        ) as mock_credit, patch(
            "miningtaxes.tasks.MININGTAXES_TAX_ONLY_CORP_MOONS", True
        ):
            tasks.update_character.run(
                character_pk=self.main_character.pk, force_update=True, celery=False
            )
        mock_update.assert_called_once()
        mock_credit.assert_called_once()
        mock_add_corp.assert_called_once()

    def test_update_all_prices_material_update_branch(self):
        from ..models import OrePrices
        from ..tests.testdata.load_eve_sde import add_material

        add_material(45511, 16635, 3)
        OrePrices.objects.all().delete()
        OrePrices.objects.create(eve_type_id=16635, buy=1, sell=1, updated=now())

        with patch(
            "miningtaxes.tasks.PriceGroups", return_value=SimpleNamespace(groups=[])
        ), patch(
            "miningtaxes.tasks.get_bulk_prices",
            side_effect=[
                {"45511": {"buy": {"max": "11"}, "sell": {"min": "22"}}},
                {"16635": {"buy": {"max": "5"}, "sell": {"min": "6"}}},
            ],
        ):
            tasks.update_all_prices.run(force=[45511])

        self.assertEqual(OrePrices.objects.get(eve_type_id=16635).buy, 5)


class TestFinalPush(MiningTaxesBaseTestCase):
    @patch.object(
        AdminCharacter,
        "fetch_token",
        return_value=SimpleNamespace(character_name="Admin", pk=1),
    )
    def test_admin_integrity_branch_and_strings(self, mock_fetch_token):
        existing = self.admin_character.mining_obs.create(
            obs_id=9010, obs_type="structure", name="Existing", sys_name="Amamake"
        )
        mock_client = SimpleNamespace(Industry=MagicMock(), Universe=MagicMock())
        mock_client.Industry.GetCorporationCorporationIdMiningObservers.return_value.results.return_value = [
            {"observer_id": 9010, "observer_type": "structure"}
        ]
        mock_client.Universe.GetUniverseStructuresStructureId.return_value = (
            SimpleNamespace(
                result=lambda use_etag=False: {
                    "name": "Existing Structure",
                    "solar_system_id": 30002537,
                }
            )
        )
        mock_client.Industry.GetCorporationCorporationIdMiningObserversObserverId.return_value.results.return_value = [
            {
                "character_id": 1001,
                "last_updated": now().date(),
                "type_id": 45511,
                "quantity": 2,
            }
        ]

        with patch(
            "miningtaxes.models.admin.esi", new=SimpleNamespace(client=mock_client)
        ), patch.object(
            type(self.admin_character.mining_obs),
            "update_or_create",
            side_effect=IntegrityError(),
        ):
            self.admin_character.update_mining_observers()

        log = existing.mining_log.first()
        ledger = self.admin_character.corp_ledger.first()
        self.assertIn("miningObs", str(existing))
        self.assertIn("miningObs", str(log))
        self.assertIn("wallet", str(ledger))

    @patch.object(
        Character,
        "fetch_token",
        return_value=SimpleNamespace(character_name="Char", pk=1),
    )
    def test_character_update_existing_row_and_misc_methods(self, mock_fetch_token):
        row = self.main_character.mining_ledger.create(
            date=self.prev_month,
            quantity=1,
            eve_type_id=1230,
            eve_solar_system_id=30000142,
        )
        from ..models import CharacterMiningLedgerEntry

        mock_client = SimpleNamespace(Industry=MagicMock())
        mock_client.Industry.GetCharactersCharacterIdMining.return_value.results.return_value = [
            {
                "date": self.prev_month,
                "quantity": 5,
                "type_id": 1230,
                "solar_system_id": 30000142,
            }
        ]
        with patch(
            "miningtaxes.models.character.esi", new=SimpleNamespace(client=mock_client)
        ), patch.object(CharacterMiningLedgerEntry, "calc_prices") as mock_calc:
            self.main_character.update_mining_ledger()
        row.refresh_from_db()
        self.assertEqual(row.quantity, 5)
        mock_calc.assert_called_once()
        self.assertEqual(
            Character.get_esi_scopes(), ["esi-industry.read_character_mining.v1"]
        )

        eve = EveCharacter.objects.create(
            character_id=1700,
            character_name="Empty",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
            alliance_ticker="WYN",
        )
        empty = Character.objects.create(eve_character=eve)
        empty.calc_lifetime_taxes()
        empty.calc_lifetime_credits()
        self.assertEqual(empty.get_monthly_credits(), {})
        self.assertIn("ISK", str(self.main_character.tax_credits.first()))
        self.assertFalse(
            CharacterUpdateStatus(character=self.main_character).is_updating
        )
        self.assertTrue(
            CharacterUpdateStatus(character=self.main_character).has_changed(
                {"a": 1}, hash_num=2
            )
        )
        self.assertTrue(
            CharacterUpdateStatus(character=self.main_character).has_changed(
                {"a": 1}, hash_num=3
            )
        )
        self.assertIsNone(CharacterMiningLedgerEntry.objects.annotate_pricing())
        self.assertIn(str(self.main_character.pk), repr(self.main_character))
        self.assertIn(str(row.id), str(row))

    def test_view_remaining_branches(self):
        from django.http import HttpResponse
        from django.urls import reverse

        from allianceauth.tests.auth_utils import AuthUtils
        from app_utils.testing import add_character_to_user

        with patch("miningtaxes.views.render", return_value=HttpResponse("ok")):
            plain = AuthUtils.create_user("plain-basic")
            plain = AuthUtils.add_permission_to_user_by_name(
                "miningtaxes.basic_access", plain
            )
            self.client.force_login(plain)
            self.assertEqual(
                self.client.get(reverse("miningtaxes:index")).status_code, 302
            )
            request = RequestFactory().get("/")
            request.user = plain
            self.assertEqual(
                __import__("inspect").unwrap(views.launcher)(request).status_code, 200
            )
            summary_request = RequestFactory().get("/")
            summary_request.user = plain
            self.assertEqual(
                __import__("inspect")
                .unwrap(views.summary_month_json)(summary_request, plain.pk)
                .status_code,
                200,
            )

            admin_user = AuthUtils.create_user("admin-alt")
            admin_user = AuthUtils.add_permission_to_user_by_name(
                "miningtaxes.basic_access", admin_user
            )
            admin_user = AuthUtils.add_permission_to_user_by_name(
                "miningtaxes.admin_access", admin_user
            )
            add_character_to_user(
                admin_user,
                EveCharacter.objects.get(character_id=1103),
                is_main=False,
                scopes=Character.get_esi_scopes(),
            )
            Character.objects.create(
                eve_character=EveCharacter.objects.get(character_id=1103)
            )
            self.client.force_login(self.user)
            self.assertEqual(
                self.client.post(
                    reverse("miningtaxes:admin_tables"),
                    {"creditbox": "4", "userid": admin_user.pk},
                ).status_code,
                200,
            )

            self.client.force_login(self.other_user)
            self.assertEqual(
                self.client.get(
                    reverse("miningtaxes:user_mining_ledger_90day", args=[self.user.pk])
                ).status_code,
                403,
            )



import datetime as dt

from dateutil.relativedelta import relativedelta

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import User
from django.db import transaction
from django.http import (
    HttpResponse,
    HttpResponseForbidden,
    HttpResponseNotFound,
    JsonResponse,
)
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.html import format_html
from django.utils.timezone import now

# from django.views.decorators.cache import cache_page
from esi.decorators import token_required

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import humanize_number
from app_utils.logging import LoggerAddTag
from app_utils.views import bootstrap_icon_plus_name_html

from . import __title__, __version__, tasks
from .forms import SettingsForm
from .models import AdminCharacter, AdminMiningObsLog, Character, Settings, Stats

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("miningtaxes.admin_access")
def admin_launcher(request):
    settings = Settings.load()
    if request.method == "POST":
        form = SettingsForm(request.POST)
        if form.is_valid():
            form = SettingsForm(request.POST, instance=settings)
            form.save()
            messages.success(
                request,
                format_html("Changes saved!"),
            )
    else:
        form = SettingsForm(instance=settings)

    admin_query = AdminCharacter.objects.all()
    auth_characters = list()
    for a_character in admin_query:
        eve_character = a_character.eve_character
        auth_characters.append(
            {
                "character_id": eve_character.character_id,
                "character_name": eve_character.character_name,
                "character": a_character,
                "alliance_id": eve_character.alliance_id,
                "alliance_name": eve_character.alliance_name,
                "corporation_id": eve_character.corporation_id,
                "corporation_name": eve_character.corporation_name,
            }
        )

    registered = Character.objects.all()
    auth_registered = list()
    for a_character in registered:
        eve_character = a_character.eve_character
        auth_registered.append(
            {
                "character_id": eve_character.character_id,
                "character_name": eve_character.character_name,
                "character": a_character,
                "alliance_id": eve_character.alliance_id,
                "alliance_name": eve_character.alliance_name,
                "corporation_id": eve_character.corporation_id,
                "corporation_name": eve_character.corporation_name,
            }
        )

    context = {
        "page_title": "Admin Settings",
        "auth_characters": auth_characters,
        "has_registered_characters": len(auth_characters) > 0,
        "auth_registered": auth_registered,
        "version": __version__,
        "form": form,
    }
    return render(request, "miningtaxes/admin_launcher.html", context)


@login_required
@permission_required("miningtaxes.admin_access")
def admin_launcher_tax_table(request):
    from .helpers import PriceGroups
    from .models import OrePrices

    data = []
    pg = PriceGroups()
    for ore in OrePrices.objects.all():
        if "Compressed" in ore.eve_type.name:
            continue
        if ore.eve_type.group_id not in pg.taxgroups:
            continue
        tax_rate = ore.tax_rate
        group = pg.taxgroups[ore.eve_type.group_id]
        data.append(
            {
                "group": group,
                "name": ore.eve_type.name,
                "tid": ore.id,
                "tax_rate": tax_rate,
            }
        )
    return JsonResponse({"data": data})


@login_required
@permission_required("miningtaxes.admin_access")
def admin_launcher_save_rates(request):
    import json

    from .models import OrePrices

    s = Stats.load()
    data = json.loads(request.POST["tax_data"])
    for d in data:
        ore = OrePrices.objects.get(id=d["tid"])
        ore.tax_rate = d["tax_rate"]
        ore.save()
    messages.info(
        request,
        format_html("Tax rates updated!"),
    )
    s.calc_ore_prices_json()
    return redirect("miningtaxes:admin_launcher")


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_char_json(request):
    s = Stats.load()
    return JsonResponse({"data": s.get_admin_char_json()})


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_main_json(request):
    s = Stats.load()
    return JsonResponse({"data": s.get_admin_main_json()})


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_tables(request):
    if request.method == "POST":
        isk = request.POST["creditbox"].replace(",", "")
        try:
            isk = float(isk)
        except ValueError:
            isk = None
            pass
        if isk is None:
            messages.warning(
                request,
                format_html("Invalid amount. Please enter a valid number"),
            )
        else:
            user = User.objects.get(pk=int(request.POST["userid"]))
            characters = Character.objects.owned_by_user(user)
            suitable = None
            for c in characters:
                if c.is_main:
                    suitable = c
                    break
                suitable = c
            suitable.give_credit(isk, "credit")
            s = Stats.load()
            s.calc_admin_main_json()
            messages.warning(
                request,
                format_html("Tax credit given!"),
            )

    context = {
        "page_title": "Admin Tables",
    }
    return render(request, "miningtaxes/admin_tables.html", context)


@login_required
@permission_required("miningtaxes.basic_access")
def ore_prices(request):
    return render(request, "miningtaxes/ore_prices.html", {})


@login_required
@permission_required("miningtaxes.basic_access")
def ore_prices_json(request):
    s = Stats.load()
    return JsonResponse({"data": s.get_ore_prices_json()})


@login_required
@permission_required("miningtaxes.basic_access")
def faq(request):
    corps = []
    for c in AdminCharacter.objects.all():
        corps.append(c.eve_character.corporation_name)
    corps = ",".join(corps)
    settings = Settings.load()
    context = {"phrase": settings.phrase, "corps": corps}
    return render(request, "miningtaxes/faq.html", context)


@login_required
@permission_required("miningtaxes.basic_access")
def index(request):
    characters = Character.objects.owned_by_user(request.user)
    if len(characters) == 0:
        return redirect("miningtaxes:launcher")
    return redirect("miningtaxes:user_summary", request.user.pk)


@login_required
@permission_required("miningtaxes.basic_access")
def user_summary(request, user_pk: int):
    s = Stats.load()
    user = User.objects.get(pk=user_pk)
    if not (
        request.user == user or request.user.has_perm("miningtaxes.auditor_access")
    ):
        return HttpResponseForbidden()
    owned_chars_query = (
        EveCharacter.objects.filter(character_ownership__user=user)
        .select_related(
            "miningtaxes_character",
        )
        .order_by("character_name")
    )
    auth_characters = list()
    unregistered_chars = list()
    for eve_character in owned_chars_query:
        try:
            character = eve_character.miningtaxes_character
        except AttributeError:
            unregistered_chars.append(eve_character.character_name)
        else:
            auth_characters.append(character)
    unregistered_chars = sorted(unregistered_chars)
    main_character_id = user.profile.main_character.character_id
    main_data, _, user2taxes = s.main_data_helper(auth_characters)
    taxes_due = user2taxes[user][0]
    if taxes_due < 0:
        taxes_due = 0

    context = {
        "page_title": "Taxes Summary",
        "auth_characters": auth_characters,
        "unregistered_chars": unregistered_chars,
        "main_character_id": main_character_id,
        "balance": humanize_number(main_data[list(main_data.keys())[0]]["balance"]),
        "balance_raw": main_data[list(main_data.keys())[0]]["balance"],
        "taxes_due": taxes_due,
        "last_paid": main_data[list(main_data.keys())[0]]["last_paid"],
        "user_pk": user_pk,
    }
    return render(request, "miningtaxes/user_summary.html", context)


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_get_all_activity_json(request):
    s = Stats.load()
    return JsonResponse({"data": s.get_admin_get_all_activity_json()})


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_corp_ledger(request):
    s = Stats.load()
    return JsonResponse(s.get_admin_corp_ledger())


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_corp_mining_history(request):
    s = Stats.load()
    return JsonResponse(s.get_admin_corp_mining_history())


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_mining_by_sys_json(request):
    s = Stats.load()
    return JsonResponse(s.get_admin_mining_by_sys_json())


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_tax_revenue_json(request):
    s = Stats.load()
    return JsonResponse(s.get_admin_tax_revenue_json())


@login_required
@permission_required("miningtaxes.auditor_access")
def admin_month_json(request):
    s = Stats.load()
    return JsonResponse(s.get_admin_month_json())


@login_required
@permission_required("miningtaxes.basic_access")
def summary_month_json(request, user_pk: int):
    user = User.objects.get(pk=user_pk)
    if request.user != user and not request.user.has_perm("miningtaxes.auditor_access"):
        return HttpResponseForbidden()
    characters = Character.objects.owned_by_user(user)
    monthly = list(map(lambda x: x.get_monthly_taxes(), characters))
    firstmonth = None
    for entries in monthly:
        if len(entries.keys()) == 0:
            continue
        if firstmonth is None or firstmonth > sorted(entries.keys())[0]:
            firstmonth = sorted(entries.keys())[0]
    xs = None
    ys = []
    for i, entries in enumerate(monthly):
        y = [characters[i].name]
        x = ["x"]
        curmonth = firstmonth
        if curmonth is None:
            continue
        lastmonth = dt.date(now().year, now().month, 1)
        while curmonth <= lastmonth:
            if curmonth not in entries:
                entries[curmonth] = 0.0
            x.append(curmonth)
            curmonth += relativedelta(months=1)

        if xs is None:
            xs = x

        for i in range(1, len(xs)):
            y.append(entries[xs[i]])
        ys.append(y)
    return JsonResponse({"xdata": xs, "ydata": ys})


@login_required
@permission_required("miningtaxes.basic_access")
def all_tax_credits(request, user_pk: int):
    user = User.objects.get(pk=user_pk)
    if request.user != user and not request.user.has_perm("miningtaxes.auditor_access"):
        return HttpResponseForbidden()
    characters = Character.objects.owned_by_user(user)
    allcredits = []
    for c in characters:
        if c.eve_character is None:
            continue
        allcredits += map(
            lambda x: {
                "date": x.date,
                "character": bootstrap_icon_plus_name_html(
                    icon_url=c.eve_character.portrait_url(),
                    name=c.eve_character.character_name,
                    size=16,
                ),
                "amount": x.credit,
                "reason": x.credit_type,
            },
            c.tax_credits.all(),
        )

    return JsonResponse({"data": allcredits})


@login_required
@permission_required("miningtaxes.basic_access")
def leaderboards(request):
    s = Stats.load()
    return JsonResponse(s.get_leaderboards())


@login_required
@permission_required("miningtaxes.basic_access")
def launcher(request) -> HttpResponse:
    owned_chars_query = (
        EveCharacter.objects.filter(character_ownership__user=request.user)
        .select_related(
            "miningtaxes_character",
        )
        .order_by("character_name")
    )
    has_auth_characters = owned_chars_query.exists()
    auth_characters = list()
    unregistered_chars = list()
    for eve_character in owned_chars_query:
        try:
            character = eve_character.miningtaxes_character
        except AttributeError:
            unregistered_chars.append(eve_character.character_name)
        else:
            auth_characters.append(
                {
                    "character_id": eve_character.character_id,
                    "character_name": eve_character.character_name,
                    "character": character,
                    "alliance_id": eve_character.alliance_id,
                    "alliance_name": eve_character.alliance_name,
                    "corporation_id": eve_character.corporation_id,
                    "corporation_name": eve_character.corporation_name,
                }
            )

    unregistered_chars = sorted(unregistered_chars)

    try:
        main_character_id = request.user.profile.main_character.character_id
    except AttributeError:
        main_character_id = None

    context = {
        "page_title": "My Characters",
        "auth_characters": auth_characters,
        "has_auth_characters": has_auth_characters,
        "unregistered_chars": unregistered_chars,
        "has_registered_characters": len(auth_characters) > 0,
        "main_character_id": main_character_id,
    }

    """
    if has_auth_characters:
        messages.warning(
            request,
            format_html(
                "Please register all your characters. "
                "You currently have <strong>{}</strong> unregistered characters.",
                unregistered_chars,
            ),
        )
    """
    return render(request, "miningtaxes/launcher.html", context)


@login_required
@permission_required("miningtaxes.admin_access")
@token_required(scopes=AdminCharacter.get_esi_scopes())
def add_admin_character(request, token) -> HttpResponse:
    eve_character = get_object_or_404(EveCharacter, character_id=token.character_id)
    with transaction.atomic():
        character, _ = AdminCharacter.objects.update_or_create(
            eve_character=eve_character
        )
    tasks.update_admin_character.delay(character_pk=character.pk)
    messages.success(
        request,
        format_html(
            "<strong>{}</strong> has been registered. "
            "Note that it can take a minute until all character data is visible.",
            eve_character,
        ),
    )
    return redirect("miningtaxes:admin_launcher")


@login_required
@permission_required("miningtaxes.basic_access")
@token_required(scopes=Character.get_esi_scopes())
def add_character(request, token) -> HttpResponse:
    eve_character = get_object_or_404(EveCharacter, character_id=token.character_id)
    with transaction.atomic():
        character, _ = Character.objects.update_or_create(eve_character=eve_character)
    tasks.update_character.delay(character_pk=character.pk)
    messages.success(
        request,
        format_html(
            "<strong>{}</strong> has been registered. "
            "Note that it can take a minute until all character data is visible.",
            eve_character,
        ),
    )
    s = Stats.load()
    s.calc_admin_main_json()
    return redirect("miningtaxes:launcher")


@login_required
@permission_required("miningtaxes.admin_access")
def purge_old_corphistory(request) -> HttpResponse:
    days_90 = now() - dt.timedelta(days=90)

    AdminMiningObsLog.objects.filter(date__lte=days_90).delete()

    messages.success(
        request,
        format_html("Purged old corp mining history as requested."),
    )
    return redirect("miningtaxes:admin_launcher")


@login_required
@permission_required("miningtaxes.admin_access")
def remove_admin_registered(request, character_pk: int) -> HttpResponse:
    try:
        character = Character.objects.select_related(
            "eve_character__character_ownership__user", "eve_character"
        ).get(pk=character_pk)
    except Character.DoesNotExist:
        return HttpResponseNotFound(f"Character with pk {character_pk} not found")

    character_name = character.eve_character.character_name

    character.delete()
    messages.success(
        request,
        format_html(
            "Removed character <strong>{}</strong> as requested.", character_name
        ),
    )
    return redirect("miningtaxes:admin_launcher")


@login_required
@permission_required("miningtaxes.admin_access")
def remove_admin_character(request, character_pk: int) -> HttpResponse:
    try:
        character = AdminCharacter.objects.select_related(
            "eve_character__character_ownership__user", "eve_character"
        ).get(pk=character_pk)
    except AdminCharacter.DoesNotExist:
        return HttpResponseNotFound(f"Character with pk {character_pk} not found")

    character_name = character.eve_character.character_name

    character.delete()
    messages.success(
        request,
        format_html(
            "Removed character <strong>{}</strong> as requested.", character_name
        ),
    )
    return redirect("miningtaxes:admin_launcher")


@login_required
@permission_required("miningtaxes.basic_access")
def remove_character(request, character_pk: int) -> HttpResponse:
    try:
        character = Character.objects.select_related(
            "eve_character__character_ownership__user", "eve_character"
        ).get(pk=character_pk)
    except Character.DoesNotExist:
        return HttpResponseNotFound(f"Character with pk {character_pk} not found")
    if character.user and character.user == request.user:
        character_name = character.eve_character.character_name

        # Notify that character has been dropped
        # permission_to_notify = Permission.objects.select_related("content_type").get(
        #    content_type__app_label=Character._meta.app_label,
        #    codename="notified_on_character_removal",
        # )
        # title = f"{__title__}: Character has been removed!"
        # message = f"{request.user} has removed character '{character_name}'"
        # for to_notify in users_with_permission(permission_to_notify):
        #    if character.user_has_scope(to_notify):
        #        notify(user=to_notify, title=title, message=message, level="INFO")

        character.delete()
        messages.success(
            request,
            format_html(
                "Removed character <strong>{}</strong> as requested.", character_name
            ),
        )
    else:
        return HttpResponseForbidden(
            f"No permission to remove Character with pk {character_pk}"
        )
    return redirect("miningtaxes:launcher")


@login_required
@permission_required("miningtaxes.basic_access")
def character_viewer(request, character_pk: int):
    character = Character.objects.get(pk=character_pk)
    context = {
        "character": character,
    }
    return render(request, "miningtaxes/character_viewer.html", context)


@login_required
@permission_required("miningtaxes.basic_access")
def user_ledger(request, user_pk: int):
    user = User.objects.get(pk=user_pk)
    if not (
        request.user == user or request.user.has_perm("miningtaxes.auditor_access")
    ):
        return HttpResponseForbidden()
    return render(request, "miningtaxes/user_ledger.html", {"user_pk": user_pk})


@login_required
@permission_required("miningtaxes.basic_access")
def user_ledger_data(request, user_pk: int):
    user = User.objects.get(pk=user_pk)
    if not (
        request.user == user or request.user.has_perm("miningtaxes.auditor_access")
    ):
        return HttpResponseForbidden()
    owned_chars_query = (
        EveCharacter.objects.filter(character_ownership__user=user)
        .select_related(
            "miningtaxes_character",
        )
        .order_by("character_name")
    )
    auth_characters = list()
    unregistered_chars = list()
    for eve_character in owned_chars_query:
        try:
            character = eve_character.miningtaxes_character
        except AttributeError:
            unregistered_chars.append(eve_character.character_name)
        else:
            auth_characters.append(character)

    combined_ledgers = {}
    for character in auth_characters:
        if request.user != character.user and not request.user.has_perm(
            "miningtaxes.auditor_access"
        ):
            return HttpResponseForbidden()
        qs = character.mining_ledger.select_related(
            "eve_solar_system",
            "eve_solar_system__constellation__region",
            "eve_type",
        )
        for row in qs:
            key = f"{row.date.isoformat()}||{row.eve_solar_system.name}||{row.eve_type.name}"
            if key not in combined_ledgers:
                combined_ledgers[key] = {
                    "q": 0,
                    "raw": 0.0,
                    "ref": 0.0,
                    "taxed": 0.0,
                    "tax": 0.0,
                }
            combined_ledgers[key]["q"] += row.quantity
            combined_ledgers[key]["raw"] += row.raw_price
            combined_ledgers[key]["ref"] += row.refined_price
            combined_ledgers[key]["taxed"] += row.taxed_value
            combined_ledgers[key]["tax"] += row.taxes_owed

    data = []
    for k in combined_ledgers:
        date, system, ore = k.split("||")
        data.append(
            {
                "date": date,
                "system": system,
                "ore": ore,
                "quantity": combined_ledgers[k]["q"],
                "raw": combined_ledgers[k]["raw"],
                "ref": combined_ledgers[k]["ref"],
                "taxed": combined_ledgers[k]["taxed"],
                "tax": combined_ledgers[k]["tax"],
            }
        )
    return JsonResponse({"data": data})


@login_required
@permission_required("miningtaxes.basic_access")
def char_mining_ledger_data(request, character_pk: int) -> JsonResponse:
    character = Character.objects.get(pk=character_pk)
    if request.user != character.user and not request.user.has_perm(
        "miningtaxes.auditor_access"
    ):
        return HttpResponseForbidden()
    qs = character.mining_ledger.select_related(
        "eve_solar_system",
        "eve_solar_system__constellation__region",
        "eve_type",
    )
    data = [
        {
            "date": row.date.isoformat(),
            "quantity": row.quantity,
            "region": row.eve_solar_system.constellation.region.name,
            "solar_system": row.eve_solar_system.name,
            "raw price": row.raw_price,
            "refined price": row.refined_price,
            "taxed value": row.taxed_value,
            "taxes owed": row.taxes_owed,
            "type": row.eve_type.name,
        }
        for row in qs
    ]
    return JsonResponse({"data": data})


@login_required
@permission_required("miningtaxes.basic_access")
def curmonthgraph(request) -> JsonResponse:
    s = Stats.load()
    dat = s.get_curmonth_leadergraph()
    return JsonResponse(dat)


@login_required
@permission_required("miningtaxes.basic_access")
def user_mining_ledger_90day(request, user_pk: int) -> JsonResponse:
    user = User.objects.get(pk=user_pk)
    if request.user != user and not request.user.has_perm("miningtaxes.auditor_access"):
        return HttpResponseForbidden()
    s = Stats.load()
    data = s.get_user_mining_ledger_90day()
    if str(user_pk) not in data:
        return HttpResponseForbidden()
    return JsonResponse(data[str(user_pk)])



import requests
from tqdm import tqdm

from django.core.management.base import BaseCommand

from ...helpers import PriceGroups
from ...models import OrePrices, Stats
from ...tasks import update_all_prices


class Command(BaseCommand):
    help = "Uses EVERef to look for all ore groups and ore children"

    def handle(self, *args, **options):
        baseurl = "https://ref-data.everef.net"
        tids = []
        newtids = []
        print("Gathering info on groups")
        for g in tqdm(PriceGroups.groups):
            r = requests.get(f"{baseurl}/groups/{g}").json()
            tids += r["type_ids"]

        print("Gathering info on all typeIDs")
        for t in tqdm(tids):
            r = requests.get(f"{baseurl}/types/{t}").json()
            if "Compressed " in r["name"]["en"]:
                continue
            try:
                OrePrices.objects.get(eve_type_id=t)
            except OrePrices.DoesNotExist:
                newtids.append(t)
                pass
        print(f"Found {len(newtids)} to force add")
        print("Running update")
        update_all_prices(force=newtids)
        print("Updating Ore Chart")
        s = Stats.load()
        s.calc_ore_prices_json()



from unittest.mock import patch

from eve_sde.models import ItemType

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from ...models import OrePrices, get_price, get_tax, ore_calc_prices
from ..testdata.load_eve_sde import add_material, load_eve_sde


class TestOrePrice(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eve_sde()

    def test_calc_prices_without_materials(self):
        ore = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=now())
        ore.calc_prices()

        self.assertEqual(ore.raw_price, 10)
        self.assertEqual(ore.refined_price, 10)
        self.assertEqual(ore.taxed_price, 10)

    def test_calc_prices_uses_refined_value_when_higher(self):
        add_material(45511, 16635, 3)
        OrePrices.objects.create(eve_type_id=16635, buy=2000, sell=200, updated=now())

        ore = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=now())
        ore.calc_prices()

        self.assertEqual(ore.raw_price, 10)
        self.assertEqual(ore.refined_price, 54.378)
        self.assertEqual(ore.taxed_price, 54.378)

    def test_ore_calc_prices_uses_raw_value_when_higher(self):
        add_material(45511, 16635, 3)
        OrePrices.objects.create(eve_type_id=16635, buy=20, sell=200, updated=now())
        ore = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=now())
        ore.calc_prices()

        prices = ore_calc_prices(ItemType.objects.get(id=45511), 20)

        self.assertEqual(prices[0], 200)
        self.assertEqual(prices[1], 10.8756)
        self.assertEqual(prices[2], 200)

    @patch("miningtaxes.models.orePrices.MININGTAXES_ALWAYS_TAX_REFINED", True)
    def test_ore_calc_prices_can_force_refined_tax_value(self):
        add_material(45511, 16635, 3)
        OrePrices.objects.create(eve_type_id=16635, buy=20, sell=200, updated=now())

        prices = ore_calc_prices(ItemType.objects.get(id=45511), 20)

        self.assertEqual(prices[0], 200)
        self.assertEqual(prices[1], 10.8756)
        self.assertEqual(prices[2], 10.8756)

    def test_get_price_falls_back_to_base_price(self):
        self.assertEqual(get_price(ItemType.objects.get(id=1230)), 5)

    def test_get_tax_returns_per_ore_rate(self):
        ore = OrePrices.objects.create(
            eve_type_id=45511, buy=10, sell=100, tax_rate=17.5, updated=now()
        )

        self.assertEqual(get_tax(ore.eve_type), 0.175)

    def test_get_tax_uses_unknown_default_when_missing(self):
        self.assertEqual(get_tax(ItemType.objects.get(id=1230)), 0.10)



# Generated by Django 4.0.7 on 2022-11-22 12:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("miningtaxes", "0005_oreprices_raw_price_oreprices_refined_price_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="charactertaxcredits",
            name="credit_type",
            field=models.CharField(default="", max_length=32),
        ),
    ]



from django.contrib.auth.models import User
from django.core.management.base import BaseCommand

from ...models import Character, Stats


class Command(BaseCommand):
    help = "Zeros all taxes of all characters"

    def handle(self, *args, **options):
        s = Stats.load()
        s.calc_admin_main_json()
        data = s.get_admin_main_json()
        for d in data:
            user = User.objects.get(pk=d["user"])
            characters = Character.objects.owned_by_user(user)
            suitable = None
            for c in characters:
                if c.is_main:
                    suitable = c
                    break
                suitable = c
            suitable.give_credit(d["balance"], "credit")
        s.calc_admin_main_json()



# Generated by Django 4.2.29 on 2026-03-21 21:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eve_sde", "0018_blueprintactivity_blueprintactivityproduct_and_more"),
        ("miningtaxes", "0013_refactor_tax_rates"),
    ]

    operations = [
        migrations.AlterField(
            model_name="adminminingobslog",
            name="eve_solar_system",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="adminminingobslog",
            name="eve_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="characterminingledgerentry",
            name="eve_solar_system",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eve_sde.solarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="characterminingledgerentry",
            name="eve_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eve_sde.itemtype",
            ),
        ),
        migrations.AlterField(
            model_name="oreprices",
            name="eve_type",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE, to="eve_sde.itemtype"
            ),
        ),
    ]
